from mage_ai.api.presenters.BasePresenter import BasePresenter


class GlobalDataProductPresenter(BasePresenter):
    default_attributes = [
        'object_type',
        'object_uuid',
        'outdated_after',
        'outdated_starting_at',
        'settings',
        'uuid',
    ]

    def present(self, **kwargs):
        return self.model.to_dict(include_uuid=True)



from typing import Union

from mage_ai.data_preparation.git import Git
from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.sync import GitConfig
from mage_ai.data_preparation.sync.base_sync import BaseSync
from mage_ai.shared.logger import VerboseFunctionExec


class GitSync(BaseSync):
    def __init__(self, sync_config: GitConfig, setup_repo: bool = True):
        self.branch = sync_config.branch or 'main'
        self.remote_repo_link = sync_config.remote_repo_link
        self.git_manager = Git(sync_config, setup_repo=setup_repo)

    def sync_data(self):
        self.git_manager.reset(self.branch)

    # Reset git sync by cloning the remote repo
    def reset(self):
        with VerboseFunctionExec(
            f'Attempting to clone from remote repo {self.remote_repo_link}',
            verbose=True,
        ):
            self.git_manager.clone()


def get_sync_config() -> Union[GitConfig, None]:
    sync_config = None
    preferences = get_preferences()

    if preferences.sync_config:
        sync_config = GitConfig.load(config=preferences.sync_config)
    return sync_config



import uuid

from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.project import Project
from mage_ai.data_preparation.repo_manager import get_repo_config
from mage_ai.orchestration.db import safe_db_query
from mage_ai.shared.hash import merge_dict
from mage_ai.usage_statistics.logger import UsageStatisticLogger


async def build_project(repo_config=None, **kwargs):
    project = Project(repo_config=repo_config)

    model = merge_dict(project.repo_config.to_dict(), dict(
        features=project.features,
        latest_version=await project.latest_version(),
        name=project.name,
        project_uuid=project.project_uuid,
        version=project.version,
    ))

    if kwargs:
        model.update(kwargs)

    return model


class ProjectResource(GenericResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        project = await self.member(None, user, **kwargs)
        collection = [project.model]

        return self.build_result_set(
            collection,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    async def member(self, _, user, **kwargs):
        model = await build_project()
        return self(model, user, **kwargs)

    @safe_db_query
    async def update(self, payload, **kwargs):
        repo_config = get_repo_config()

        data = {}
        should_log_project = self.model.get('help_improve_mage') or False

        if 'features' in payload:
            for k, v in payload.get('features', {}).items():
                feature = (repo_config.features or {}).get(k)
                if (feature and not v) or (not feature and v):
                    if 'features' not in data:
                        data['features'] = {}
                    data['features'][k] = v

        if 'help_improve_mage' in payload:
            if payload['help_improve_mage']:
                should_log_project = True

                if not repo_config.project_uuid:
                    data['project_uuid'] = uuid.uuid4().hex

            data['help_improve_mage'] = payload['help_improve_mage']

        if 'openai_api_key' in payload:
            openai_api_key = payload.get('openai_api_key')
            if repo_config.openai_api_key != openai_api_key:
                data['openai_api_key'] = payload.get('openai_api_key')

        if len(data.keys()) >= 1:
            repo_config.save(**data)

        self.model = await build_project(repo_config)

        if should_log_project:
            await UsageStatisticLogger().project_impression()

        return self



import json
import os
from typing import Dict, List

import boto3
from botocore.config import Config

from mage_ai.services.aws.ecs.config import EcsConfig


def run_task(
    command: str,
    ecs_config: EcsConfig,
    wait_for_completion: bool = True,
) -> None:
    if type(ecs_config) is dict:
        ecs_config = EcsConfig.load(config=ecs_config)
    client = boto3.client('ecs')
    response = client.run_task(**ecs_config.get_task_config(command=command))

    print(json.dumps(response, indent=4, default=str))

    if wait_for_completion:
        arn = response['tasks'][0]['taskArn']
        waiter = client.get_waiter('tasks_stopped')
        wait_kwargs = dict(
            cluster=ecs_config.cluster,
            tasks=[arn],
        )
        if ecs_config.wait_timeout:
            wait_kwargs['WaiterConfig'] = {
                'Delay': 15,
                'MaxAttempts': int(ecs_config.wait_timeout / 15),
            }
        waiter.wait(**wait_kwargs)

        tasks = client.describe_tasks(
            cluster=ecs_config.cluster,
            tasks=[arn]
        ).get('tasks')

        if not tasks:
            raise Exception('Failed to get ECS task status.')

        containers = [c for c in tasks[0]['containers'] if c['name'] == ecs_config.container_name]
        if not containers:
            raise Exception(f'Failed to get status from container {ecs_config.container_name}')

        exit_code = containers[0]['exitCode']
        if exit_code != 0:
            raise Exception(f'Container {ecs_config.container_name}'
                            f' returns non-zero exit code: {exit_code}')

    return response


def stop_task(task_arn: str, cluster: str = None) -> None:
    client = boto3.client('ecs')
    return client.stop_task(
        cluster=cluster,
        task=task_arn,
    )


def list_tasks(cluster) -> List[Dict]:
    region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
    config = Config(region_name=region_name)
    ecs_client = boto3.client('ecs', config=config)

    task_arns = ecs_client.list_tasks(
        cluster=cluster,
    )['taskArns']

    return ecs_client.describe_tasks(
        cluster=cluster,
        include=[
            'TAGS',
        ],
        tasks=task_arns,
    )['tasks']


def list_services(cluster) -> List[Dict]:
    region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
    config = Config(region_name=region_name)
    ecs_client = boto3.client('ecs', config=config)

    service_arns = ecs_client.list_services(
        cluster=cluster,
    )['serviceArns']

    return ecs_client.describe_services(
        cluster=cluster,
        include=[
            'TAGS',
        ],
        services=service_arns,
    )['services']



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.repo_manager import get_project_uuid
from mage_ai.orchestration.constants import Entity


class LlmPolicy(BasePolicy):
    @property
    def entity(self):
        parent_model = self.options.get('parent_model')
        if parent_model:
            if issubclass(parent_model.__class__, Pipeline):
                return Entity.PIPELINE, parent_model.uuid

        return Entity.PROJECT, get_project_uuid()


LlmPolicy.allow_actions([
    constants.CREATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role())


LlmPolicy.allow_read([
    'response',
    'use_case',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role())


LlmPolicy.allow_write([
    'request',
    'use_case',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role())



import enum
import os
from dataclasses import dataclass, field
from typing import Dict, List

import yaml

from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.array import find
from mage_ai.shared.hash import index_by
from mage_ai.shared.io import safe_write
from mage_ai.shared.utils import clean_name


class GlobalDataProductObjectType(str, enum.Enum):
    BLOCK = 'block'
    PIPELINE = 'pipeline'


@dataclass
class GlobalDataProduct:
    uuid: str
    object_type: str = None
    object_uuid: str = None
    outdated_after: Dict = field(default_factory=dict)
    outdated_starting_at: Dict = field(default_factory=dict)
    settings: Dict = field(default_factory=dict)

    def __init__(self, uuid: str, **kwargs):
        self.object_type = kwargs.get('object_type')
        self.object_uuid = kwargs.get('object_uuid')
        self.outdated_after = kwargs.get('outdated_after')
        self.outdated_starting_at = kwargs.get('outdated_starting_at')
        self.settings = kwargs.get('settings')
        self.uuid = clean_name(uuid) if uuid else uuid

    @classmethod
    def file_path(self) -> str:
        return os.path.join(get_repo_path(), 'global_data_products.yaml')

    @classmethod
    def load_all(self) -> List['GlobalDataProduct']:
        arr = []

        if not os.path.exists(self.file_path()):
            return []

        with open(self.file_path()) as fp:
            content = fp.read()
            if content:
                yaml_config = yaml.safe_load(content) or {}
                for uuid, config_attributes in yaml_config.items():
                    arr.append(self(
                        uuid,
                        **config_attributes,
                    ))

        return arr

    @classmethod
    def get(self, uuid: str):
        return find(lambda x: x.uuid == uuid, self.load_all())

    def to_dict(self, include_uuid: bool = False) -> Dict:
        data = dict(
            object_type=self.object_type,
            object_uuid=self.object_uuid,
            outdated_after=self.outdated_after,
            outdated_starting_at=self.outdated_starting_at,
            settings=self.settings,
        )

        if include_uuid:
            data['uuid'] = self.uuid

        return data

    def delete(self) -> None:
        self.save(delete=True)

    def save(self, delete: bool = False) -> None:
        mapping = index_by(lambda x: x.uuid, self.load_all())

        if delete:
            mapping.pop(self.uuid, None)
        else:
            mapping[self.uuid] = self

        mapping_new = {}

        items = sorted(list(mapping.items()), key=lambda tup: tup[0])
        for tup in items:
            uuid, gdp = tup
            mapping_new[uuid] = {}
            for k, v in gdp.to_dict().items():
                if v is not None:
                    mapping_new[uuid][k] = v

        content = yaml.safe_dump(mapping_new)
        safe_write(self.file_path(), content)

    def update(self, payload: Dict) -> None:
        for key in [
            'object_type',
            'object_uuid',
            'outdated_after',
            'outdated_starting_at',
            'settings',
        ]:
            value = payload.get(key)
            if getattr(self, key) != value:
                setattr(self, key, value)

        uuid = payload.get('uuid')
        if self.uuid != uuid:
            self.delete()
            self.uuid = uuid

        self.save()



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.orchestration.db.models.schedules import PipelineSchedule
from mage_ai.data_preparation.models.triggers import (
    Trigger,
    add_or_update_trigger_for_pipeline_and_persist,
    get_triggers_by_pipeline,
)
from mage_ai.orchestration.db import safe_db_query


class PipelineTriggerResource(GenericResource):
    @classmethod
    @safe_db_query
    def collection(self, query, meta, user, **kwargs):
        parent_model = kwargs['parent_model']

        return self.build_result_set(
            get_triggers_by_pipeline(parent_model.uuid),
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        parent_model = kwargs['parent_model']
        error = ApiError.RESOURCE_INVALID

        pipeline_schedule_id = payload.get('pipeline_schedule_id')
        trigger_name = payload.get('name')
        if not pipeline_schedule_id and not trigger_name:
            error.update(dict(message='Pipeline schedule ID or trigger name is required.'))
            raise ApiError(error)

        trigger = None
        if pipeline_schedule_id:
            pipeline_schedule = PipelineSchedule.query.get(pipeline_schedule_id)
            trigger = Trigger(
                name=pipeline_schedule.name,
                pipeline_uuid=pipeline_schedule.pipeline_uuid,
                schedule_interval=pipeline_schedule.schedule_interval,
                schedule_type=pipeline_schedule.schedule_type,
                settings=pipeline_schedule.settings,
                sla=pipeline_schedule.sla,
                start_time=pipeline_schedule.start_time,
                status=pipeline_schedule.status,
                variables=pipeline_schedule.variables,
            )
        else:
            trigger = Trigger(**payload)

        if not trigger:
            error.update(dict(message='Trigger failed to save or update.'))
            raise ApiError(error)

        add_or_update_trigger_for_pipeline_and_persist(trigger, parent_model.uuid)

        return self(trigger, user, **kwargs)



from typing import Dict

from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.cache.block_action_object.constants import (
    OBJECT_TYPE_BLOCK_FILE,
    OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE,
    OBJECT_TYPE_MAGE_TEMPLATE,
)
from mage_ai.data_preparation.models.constants import (
    BlockLanguage,
    BlockType,
    PipelineType,
)
from mage_ai.services.search.constants import SEARCH_TYPE_BLOCK_ACTION_OBJECTS


def filter_results(result: Dict) -> bool:
    block_action_object = result.get('block_action_object')
    object_type = result.get('object_type')

    block_type = None
    language = None

    if OBJECT_TYPE_BLOCK_FILE == object_type:
        block_type = block_action_object.get('type')
        language = block_action_object.get('language')
    elif OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE == object_type:
        block_type = block_action_object.get('block_type')
        language = block_action_object.get('language')
    elif OBJECT_TYPE_MAGE_TEMPLATE == object_type:
        block_type = block_action_object.get('block_type')
        language = block_action_object.get('language')

    if BlockLanguage.YAML == language and BlockType.DBT != block_type:
        return False

    if block_type in [
        BlockType.CALLBACK,
        BlockType.CONDITIONAL,
        BlockType.EXTENSION,
    ]:
        return False

    return True


class SearchResultResource(GenericResource):
    @classmethod
    async def create(self, payload: Dict, user, **kwargs):
        pipeline_type = payload.get('pipeline_type', None)
        query = payload.get('query', None)
        ratio = payload.get('ratio', None)
        search_type = payload.get('type', None)

        results = []

        if query:
            if SEARCH_TYPE_BLOCK_ACTION_OBJECTS == search_type:
                from mage_ai.services.search.block_action_objects import search

                results = await search(query, ratio=ratio)

                # TODO (tommy dangerous): remove this when we unify pipeline types
                if PipelineType.PYTHON == pipeline_type:
                    results = list(filter(filter_results, results))

                results = results[:12]

        return self(dict(
            results=results,
            type=search_type,
            uuid=query,
        ), user, **kwargs)



import os


class SecretsManager:
    def __init__(self):
        from azure.identity import DefaultAzureCredential
        from azure.keyvault.secrets import SecretClient

        key_vault_url = os.getenv('AZURE_KEY_VAULT_URL')
        if not key_vault_url:
            raise Exception('Please provide valid AZURE_KEY_VAULT_URL in environment variable.')
        cred = DefaultAzureCredential()
        self.client = SecretClient(vault_url=key_vault_url, credential=cred)


secrets_manager = SecretsManager()


def get_secret(secret_name: str) -> str:
    return secrets_manager.client.get_secret(secret_name).value



from mage_ai.api.presenters.BasePresenter import BasePresenter


class OauthAccessTokenPresenter(BasePresenter):
    default_attributes = [
        'expires',
        'token',
        'user_id',
    ]



"""Add key uuid to secrets

Revision ID: e3593cc2191e
Revises: 90d978a8aef8
Create Date: 2023-07-14 15:53:06.580076

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e3593cc2191e'
down_revision = '90d978a8aef8'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('secret', schema=None) as batch_op:
        batch_op.drop_constraint('name_unique', type_='unique')
        batch_op.add_column(sa.Column('key_uuid', sa.String(length=255), nullable=True))
        batch_op.create_unique_constraint('name_key_uuid_uc', ['name', 'key_uuid'])
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('secret', schema=None) as batch_op:
        batch_op.drop_constraint('name_key_uuid_uc', type_='unique')
        batch_op.drop_column('key_uuid')
        batch_op.create_unique_constraint('name_unique', columns=['name'])

    # ### end Alembic commands ###



from mage_ai.api.operations import constants
from mage_ai.api.presenters.BasePresenter import BasePresenter
from mage_ai.data_preparation.models.constants import DATAFRAME_SAMPLE_COUNT_PREVIEW
from mage_ai.data_preparation.models.custom_templates.custom_block_template import (
    CustomBlockTemplate,
)
from mage_ai.data_preparation.models.custom_templates.custom_pipeline_template import (
    CustomPipelineTemplate,
)


class CustomTemplatePresenter(BasePresenter):
    default_attributes = [
        'block_type',
        'color',
        'configuration',
        'description',
        'language',
        'name',
        'pipeline',
        'tags',
        'template_uuid',
        'user',
        'uuid',
    ]

    async def present(self, **kwargs):
        data = self.resource.model.to_dict()

        if isinstance(self.resource.model, CustomBlockTemplate):
            if kwargs.get('format') in [constants.DETAIL, constants.UPDATE]:
                data['content'] = self.resource.model.load_template_content()
        elif isinstance(self.resource.model, CustomPipelineTemplate):
            if kwargs.get('format') in [constants.DETAIL]:
                pipeline = self.resource.model.build_pipeline()

                data['pipeline'] = await pipeline.to_dict_async(
                    include_block_metadata=True,
                    include_block_pipelines=True,
                    include_block_tags=True,
                    include_callback_blocks=True,
                    include_conditional_blocks=True,
                    include_content=False,
                    include_extensions=True,
                    include_outputs=False,
                    sample_count=DATAFRAME_SAMPLE_COUNT_PREVIEW,
                )

        return data


CustomTemplatePresenter.register_formats([
    constants.DETAIL,
    constants.UPDATE,
], CustomTemplatePresenter.default_attributes + [
        'content',
    ],
)



import os
from typing import Dict, List

from mage_ai.data_preparation.models.custom_templates.constants import (
    CUSTOM_TEMPLATES_DIRECTORY,
    CUSTOM_TEMPLATES_DIRECTORY_ENVIRONMENT_VARIABLE,
)
from mage_ai.data_preparation.models.file import File
from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.hash import group_by


def custom_templates_directory() -> str:
    return os.getenv(
        CUSTOM_TEMPLATES_DIRECTORY_ENVIRONMENT_VARIABLE,
    ) or CUSTOM_TEMPLATES_DIRECTORY


def get_templates(object_type_directory: str) -> List[Dict]:
    full_path = os.path.join(
        get_repo_path(),
        custom_templates_directory(),
        object_type_directory,
    )

    if os.path.exists(full_path):
        file_dict = File.get_all_files(full_path)
        if file_dict:
            return file_dict.get('children', [])


def flatten_files(
    children: List[Dict],
    parent_names: List[str] = None,
) -> List[Dict]:
    if not children:
        return []
    arr = []

    for child1 in children:
        children2 = child1.get('children')
        if children2:
            names = parent_names.copy() if parent_names else []
            if child1.get('name'):
                names.append(child1.get('name'))
            arr += flatten_files(children2, names)
        else:
            child_updated = child1.copy()
            child_updated['parent_names'] = parent_names
            arr.append(child_updated)

    return arr


def group_and_hydrate_files(
    file_dicts: List[Dict],
    custom_template_class,
) -> List:
    groups = group_by(lambda x: os.path.join(*x.get('parent_names', [])), file_dicts)

    arr = []

    for template_uuid, _ in groups.items():
        custom_template = custom_template_class.load(template_uuid=template_uuid)
        if custom_template:
            arr.append(custom_template)

    return arr



from mage_ai.api.presenters.BasePresenter import BasePresenter


class OauthPresenter(BasePresenter):
    default_attributes = [
        'authenticated',
        'expires',
        'provider',
        'url',
    ]



import os
import shutil
import yaml
from dataclasses import dataclass, field
from jinja2 import Template
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.constants import (
    BLOCK_LANGUAGE_TO_FILE_EXTENSION,
    BlockColor,
    BlockLanguage,
    BlockType,
)
from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.custom_templates.constants import (
    DIRECTORY_FOR_BLOCK_TEMPLATES,
    METADATA_FILENAME_WITH_EXTENSION,
)
from mage_ai.data_preparation.models.custom_templates.utils import custom_templates_directory
from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.config import BaseConfig
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.io import safe_write
from typing import Dict, List


@dataclass
class CustomBlockTemplate(BaseConfig):
    block_type: BlockType = None
    color: BlockColor = None
    configuration: Dict = None
    content: str = ''
    description: str = None
    language: BlockLanguage = None
    name: str = None
    pipeline: Dict = field(default_factory=dict)
    tags: List = field(default_factory=list)
    template_uuid: str = None
    user: Dict = field(default_factory=dict)

    @classmethod
    def load(self, template_uuid: str = None, uuid: str = None):
        uuid_use = uuid
        template_uuid_use = template_uuid

        if uuid_use:
            parts = uuid_use.split(os.sep)
            template_uuid_use = os.path.join(*parts[2:])
        elif template_uuid_use:
            uuid_use = os.path.join(
                custom_templates_directory(),
                DIRECTORY_FOR_BLOCK_TEMPLATES,
                template_uuid_use,
            )

        try:
            config_path_metadata = os.path.join(
                get_repo_path(),
                uuid_use,
                METADATA_FILENAME_WITH_EXTENSION,
            )
            custom_template = super().load(config_path_metadata)
            custom_template.template_uuid = template_uuid_use

            return custom_template
        except Exception as err:
            print(f'[WARNING] CustomBlockTemplate.load: {err}')

    @property
    def uuid(self):
        return os.path.join(
            custom_templates_directory(),
            DIRECTORY_FOR_BLOCK_TEMPLATES,
            self.template_uuid,
        )

    @property
    def metadata_file_path(self) -> str:
        return os.path.join(
            get_repo_path(),
            self.uuid,
            METADATA_FILENAME_WITH_EXTENSION,
        )

    def create_block(
        self,
        block_name: str,
        pipeline: Pipeline,
        extension_uuid: str = None,
        priority: int = None,
        upstream_block_uuids: List[str] = None,
        **kwargs,
    ) -> Block:
        configuration = None
        if self.configuration and type(self.configuration) is dict:
            configuration = self.configuration

        return Block.create(
            block_name,
            self.block_type,
            get_repo_path(),
            color=self.color,
            configuration=configuration,
            extension_uuid=extension_uuid,
            language=self.language,
            pipeline=pipeline,
            priority=priority,
            upstream_block_uuids=upstream_block_uuids,
        )

    def load_template_content(self, language: BlockLanguage = None) -> str:
        language_to_use = language or self.language
        filename = '.'.join([
            self.template_uuid,
            BLOCK_LANGUAGE_TO_FILE_EXTENSION.get(language_to_use, ''),
        ])

        return File(
            dir_path=self.uuid,
            filename=filename,
            repo_path=get_repo_path(),
        ).content()

    def render_template(
        self,
        language: BlockLanguage = None,
        variables: Dict = None,
    ) -> str:
        content = self.load_template_content(language)
        if content:
            return Template(content).render(**(variables or {}))

    def to_dict(self, include_content: bool = False) -> Dict:
        data = merge_dict(self.to_dict_base(), dict(
            template_uuid=self.template_uuid,
            uuid=self.uuid,
        ))

        if include_content:
            data['content'] = self.load_template_content()

        return data

    def to_dict_base(self) -> Dict:
        return dict(
            block_type=self.block_type,
            color=self.color,
            configuration=self.configuration,
            description=self.description,
            language=self.language,
            name=self.name,
            pipeline=self.pipeline,
            tags=self.tags,
            user=self.user,
        )

    def save(self) -> None:
        content = yaml.safe_dump(self.to_dict_base())
        file_path = self.metadata_file_path
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        safe_write(file_path, content)

        if self.content:
            filename = '.'.join([
                self.template_uuid,
                BLOCK_LANGUAGE_TO_FILE_EXTENSION[self.language],
            ])

            File.create(
                filename,
                self.uuid,
                self.content,
                get_repo_path(),
            )

    def delete(self) -> None:
        shutil.rmtree(os.path.join(get_repo_path(), self.uuid))



import warnings
from typing import Union

import oracledb
from pandas import DataFrame, read_sql

from mage_ai.io.base import QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.io.sql import BaseSQL


class OracleDB(BaseSQL):
    def __init__(self,
                 user,
                 password,
                 host,
                 port,
                 service_name,
                 verbose: bool = False,
                 **kwargs) -> None:
        super().__init__(user=user,
                         password=password,
                         host=host,
                         port=port,
                         service_name=service_name,
                         verbose=verbose,
                         **kwargs)

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'OracleDB':
        return cls(
            user=config[ConfigKey.ORACLEDB_USER],
            password=config[ConfigKey.ORACLEDB_PASSWORD],
            host=config[ConfigKey.ORACLEDB_HOST],
            port=config[ConfigKey.ORACLEDB_PORT],
            service_name=config[ConfigKey.ORACLEDB_SERVICE]
        )

    def open(self) -> None:
        with self.printer.print_msg('Opening connection to OracleDB database (thin mode)'):
            self._ctx = oracledb.connect(**self.settings)

    def load(
        self,
        query_string: str,
        limit: int = QUERY_ROW_LIMIT,
        display_query: Union[str, None] = None,
        verbose: bool = True,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from the connected database into a Pandas data frame based on the query given.
        This will fail if the query returns no data from the database. This function will load at
        maximum 10,000,000 rows of data. To operate on more data, consider performing data
        transformations in warehouse.

        Args:
            query_string (str): Query to execute on the database.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults
                to 10,000,000.
            **kwargs: Additional query parameters.

        Returns:
            DataFrame: The data frame corresponding to the data returned by the given query.
        """
        print_message = 'Loading data'
        if verbose:
            print_message += ' with query'

            if display_query:
                for line in display_query.split('\n'):
                    print_message += f'\n{line}'
            else:
                print_message += f'\n\n{query_string}\n\n'

        query_string = self._clean_query(query_string)

        with self.printer.print_msg(print_message):
            warnings.filterwarnings('ignore', category=UserWarning)

            return read_sql(
                self._enforce_limit_oracledb(query_string, limit),
                self.conn,
                **kwargs,
            )

    def _enforce_limit_oracledb(self, query: str, limit: int = QUERY_ROW_LIMIT) -> str:
        """
        Modifies SQL SELECT query to enforce a limit on the number of rows returned by the query.
        This method is currently supports Oracledb syntax only.

        Args:
            query (str): The SQL query to modify
            limit (int): The limit on the number of rows to return.

        Returns:
            str: Modified query with limit on row count returned.
        """
        query = query.strip(';')

        return f"""
WITH subquery AS (
    {query}
)

SELECT *
FROM subquery
FETCH FIRST {limit} ROWS ONLY
                """



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.BlockPresenter import BlockPresenter
from mage_ai.orchestration.constants import Entity


class BlockPolicy(BasePolicy):
    @property
    def entity(self):
        parent_model = self.options.get('parent_model')
        if parent_model:
            return Entity.PIPELINE, parent_model.uuid

        return super().entity


BlockPolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

BlockPolicy.allow_actions([
    constants.DETAIL,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockPolicy.allow_read([
    'content',
] + BlockPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

BlockPolicy.allow_read(BlockPresenter.default_attributes + [], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DELETE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

BlockPolicy.allow_read([
    'bookmarks',
    'content',
    'outputs',
    'pipelines',
] + BlockPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockPolicy.allow_write([
    'block_action_object',
    'color',
    'config',
    'configuration',
    'content',
    'converted_from',
    'extension_uuid',
    'has_callback',
    'language',
    'metadata',
    'name',
    'pipelines',
    'priority',
    'replicated_block',
    'type',
    'upstream_blocks',
    'uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

BlockPolicy.allow_write([
    'all_upstream_blocks_executed',
    'bookmark_values',
    'callback_blocks',
    'color',
    'conditional_blocks',
    'configuration',
    'content',
    'destination_table',
    'downstream_blocks',
    'executor_config',
    'executor_type',
    'extension_uuid',
    'has_callback',
    'language',
    'metadata',
    'name',
    'outputs',
    'pipelines',
    'replicated_block',
    'retry_config',
    'status',
    'tags',
    'tap_stream_id',
    'type',
    'upstream_blocks',
    'uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())

BlockPolicy.allow_query([
    'destination_table',
    'state_stream',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockPolicy.allow_query([
    'extension_uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DELETE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockPolicy.allow_query([
    'block_language',
    'block_type',
    'extension_uuid',
    'update_state',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_editor_role_and_pipeline_edit_access())



import asyncio
import os
from datetime import datetime
from mage_ai.cache.base import BaseCache
from mage_ai.cache.constants import CACHE_KEY_TAGS_TO_OBJECT_MAPPING
from mage_ai.cache.utils import build_pipeline_dict
from typing import Dict

KEY_FOR_PIPELINES = 'Pipeline'


class TagCache(BaseCache):
    cache_key = CACHE_KEY_TAGS_TO_OBJECT_MAPPING

    @classmethod
    async def initialize_cache(self, replace: bool = False) -> 'TagCache':
        cache = self()
        if replace or not cache.exists():
            await cache.initialize_cache_for_all_objects()

        return cache

    def get_tags(self) -> Dict:
        return self.get(self.cache_key) or {}

    def build_key(self, tag_uuid: str) -> str:
        """Generate cache key for tag.

        Args:
            tag_uuid: str: The UUID of the tag.

        Returns:
            str: The cache key.
        """
        return os.path.join(tag_uuid)

    def add_pipeline(self, tag_uuid: str, pipeline) -> None:
        self.update_pipeline(tag_uuid, pipeline, added_at=datetime.utcnow().timestamp())

    def update_pipeline(
        self,
        tag_uuid: str,
        pipeline,
        added_at: str = None,
    ) -> None:
        self.update_pipelines(tag_uuid, [pipeline], added_at)

    def update_pipelines(
        self,
        tag_uuid: str,
        pipelines,
        added_at: str = None,
    ) -> None:
        key = self.build_key(tag_uuid)
        if not key:
            return

        mapping = self.get(self.cache_key)
        if mapping is None:
            mapping = {}

        if key not in mapping:
            mapping[key] = {}

        if KEY_FOR_PIPELINES not in mapping[key]:
            mapping[key][KEY_FOR_PIPELINES] = {}

        pipelines_dict = mapping[key].get(KEY_FOR_PIPELINES, {})

        for pipeline in pipelines:
            pipeline_uuid = pipeline.get('uuid') if type(pipeline) is dict else pipeline.uuid
            pipelines_dict[pipeline_uuid] = build_pipeline_dict(
                pipeline,
                added_at=added_at,
            )

        mapping[key][KEY_FOR_PIPELINES] = pipelines_dict

        self.set(self.cache_key, mapping)

    def remove_pipeline(self, tag_uuid: str, pipeline_uuid: str) -> None:
        key = self.build_key(tag_uuid)
        if not key:
            return

        mapping = self.get(self.cache_key)
        if mapping is None:
            mapping = {}

        if key not in mapping:
            mapping[key] = {}

        pipelines_dict = mapping[key].get(KEY_FOR_PIPELINES, {})
        pipelines_dict.pop(pipeline_uuid, None)
        mapping[key][KEY_FOR_PIPELINES] = pipelines_dict

        self.set(self.cache_key, mapping)

    async def initialize_cache_for_all_objects(self) -> None:
        from mage_ai.data_preparation.models.pipeline import Pipeline

        pipeline_uuids = Pipeline.get_all_pipelines(self.repo_path)

        pipeline_dicts = await asyncio.gather(
            *[Pipeline.load_metadata(uuid, raise_exception=False) for uuid in pipeline_uuids],
        )
        pipeline_dicts = [p for p in pipeline_dicts if p is not None]

        mapping = {}

        for pipeline_dict in pipeline_dicts:
            for tag_uuid in pipeline_dict.get('tags', []):
                key = self.build_key(tag_uuid)
                if not key:
                    continue

                if key not in mapping:
                    mapping[key] = {}

                if KEY_FOR_PIPELINES not in mapping[key]:
                    mapping[key][KEY_FOR_PIPELINES] = {}

                mapping[key][KEY_FOR_PIPELINES][pipeline_dict.get('uuid')] = build_pipeline_dict(
                    pipeline_dict,
                )

        self.set(self.cache_key, mapping)



from enum import Enum


class LLMUseCase(str, Enum):
    GENERATE_COMMENT_FOR_BLOCK = 'generate_comment_for_block'
    GENERATE_DOC_FOR_BLOCK = 'generate_doc_for_block'
    GENERATE_DOC_FOR_PIPELINE = 'generate_doc_for_pipeline'
    GENERATE_BLOCK_WITH_DESCRIPTION = 'generate_block_with_description'



import urllib.parse

from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.cache.block import BlockCache
from mage_ai.cache.block_action_object import BlockActionObjectCache
from mage_ai.cache.block_action_object.constants import (
    OBJECT_TYPE_BLOCK_FILE,
    OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE,
    OBJECT_TYPE_MAGE_TEMPLATE,
)
from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.block.dbt import DBTBlock
from mage_ai.data_preparation.models.block.utils import clean_name
from mage_ai.data_preparation.models.constants import (
    FILE_EXTENSION_TO_BLOCK_LANGUAGE,
    BlockLanguage,
    BlockType,
)
from mage_ai.data_preparation.models.custom_templates.custom_block_template import (
    CustomBlockTemplate,
)
from mage_ai.data_preparation.utils.block.convert_content import convert_to_block
from mage_ai.orchestration.db import safe_db_query
from mage_ai.settings.repo import get_repo_path


class BlockResource(GenericResource):
    @classmethod
    @safe_db_query
    async def create(self, payload, user, **kwargs):
        pipeline = kwargs.get('parent_model')

        block_type = payload.get('type')
        content = payload.get('content')
        language = payload.get('language')
        name = payload.get('name')
        block_name = name or payload.get('uuid')

        payload_config = payload.get('config') or {}

        block_action_object = payload.get('block_action_object')
        if block_action_object:
            object_type = block_action_object.get('object_type')

            cache_block_action_object = await BlockActionObjectCache.initialize_cache()
            mapping = cache_block_action_object.load_all_data()
            objects_mapping = mapping.get(object_type)
            object_uuid = block_action_object.get('uuid')
            object_from_cache = objects_mapping.get(object_uuid)

            if OBJECT_TYPE_BLOCK_FILE == object_type:
                block_name = object_from_cache.get('uuid')
                block_type = object_from_cache.get('type')
                language = object_from_cache.get('language')
            elif OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE == object_type:
                payload_config['custom_template_uuid'] = object_from_cache.get('template_uuid')
            elif OBJECT_TYPE_MAGE_TEMPLATE == object_type:
                block_type = object_from_cache.get('block_type')
                language = object_from_cache.get('language')
                payload_config['template_path'] = object_from_cache.get('path')
                payload_config['template_variables'] = object_from_cache.get('template_variables')

        """
        New DBT models include "content" in its block create payload,
        whereas creating blocks from existing DBT model files do not.
        """
        if payload.get('type') == BlockType.DBT and content and language == BlockLanguage.SQL:
            dbt_block = DBTBlock(
                name,
                clean_name(name),
                BlockType.DBT,
                configuration=payload.get('configuration'),
                language=language,
            )
            if dbt_block.file_path and dbt_block.file.exists():
                raise Exception('DBT model at that folder location already exists. \
                    Please choose a different model name, or add a DBT model by \
                    selecting single model from file.')

        block_attributes = dict(
            color=payload.get('color'),
            config=payload_config,
            configuration=payload.get('configuration'),
            extension_uuid=payload.get('extension_uuid'),
            language=language,
            pipeline=pipeline,
            priority=payload.get('priority'),
            upstream_block_uuids=payload.get('upstream_blocks', []),
        )

        replicated_block_uuid = payload.get('replicated_block')
        if replicated_block_uuid:
            replicated_block = pipeline.get_block(replicated_block_uuid)
            if replicated_block:
                block_type = replicated_block.type
                block_attributes['language'] = replicated_block.language
                block_attributes['replicated_block'] = replicated_block.uuid
            else:
                error = ApiError.RESOURCE_INVALID.copy()
                error.update(
                    message=f'Replicated block {replicated_block_uuid} ' +
                    f'does not exist in pipeline {pipeline.uuid}.',
                )
                raise ApiError(error)

        if payload_config and payload_config.get('custom_template_uuid'):
            template_uuid = payload_config.get('custom_template_uuid')
            custom_template = CustomBlockTemplate.load(template_uuid=template_uuid)
            block = custom_template.create_block(
                block_name,
                pipeline,
                extension_uuid=block_attributes.get('extension_uuid'),
                priority=block_attributes.get('priority'),
                upstream_block_uuids=block_attributes.get('upstream_block_uuids'),
            )
            content = custom_template.load_template_content()
        else:
            block = Block.create(
                block_name,
                block_type,
                get_repo_path(),
                **block_attributes,
            )

        if content:
            if payload.get('converted_from'):
                content = convert_to_block(block, content)

            block.update_content(content)

        cache = await BlockCache.initialize_cache()
        cache.add_pipeline(block, pipeline)

        cache_block_action_object = await BlockActionObjectCache.initialize_cache()
        cache_block_action_object.update_block(block)

        return self(block, user, **kwargs)

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        error = ApiError.RESOURCE_INVALID.copy()

        query = kwargs.get('query', {})

        extension_uuid = query.get('extension_uuid', [None])
        if extension_uuid:
            extension_uuid = extension_uuid[0]

        block_type = query.get('block_type', [None])
        if block_type:
            block_type = block_type[0]

        pipeline = kwargs.get('parent_model')
        if pipeline:
            block = pipeline.get_block(pk, block_type=block_type, extension_uuid=extension_uuid)
            if block:
                return self(block, user, **kwargs)
            else:
                if extension_uuid:
                    message = f'Block {pk} does not exist in pipeline {pipeline.uuid} ' \
                        f'for extension {extension_uuid}.'
                else:
                    message = f'Block {pk} does not exist in pipeline {pipeline.uuid}.'
                error.update(message=message)
                raise ApiError(error)

        block_type_and_uuid = urllib.parse.unquote(pk)
        parts = block_type_and_uuid.split('/')

        if len(parts) < 2:
            error.update(message='The url path should be in block_type/block_uuid format.')
            raise ApiError(error)

        block_type = parts[0]
        block_uuid_with_extension = '/'.join(parts[1:])
        parts2 = block_uuid_with_extension.split('.')

        language = None
        if len(parts2) >= 2:
            block_uuid = '.'.join(parts2[:-1])
            language = FILE_EXTENSION_TO_BLOCK_LANGUAGE[parts2[-1]]
        else:
            block_uuid = block_uuid_with_extension

        block_language = query.get('block_language', [None])
        if block_language:
            block_language = block_language[0]
        if block_language:
            language = block_language

        if BlockType.DBT == block_type:
            block = DBTBlock(
                block_uuid,
                block_uuid,
                block_type,
                configuration=dict(file_path=block_uuid_with_extension),
                language=language,
            )
        else:
            block = Block.get_block(block_uuid, block_uuid, block_type, language=language)

        if not block.exists():
            error.update(ApiError.RESOURCE_NOT_FOUND)
            raise ApiError(error)

        return self(block, user, **kwargs)

    @safe_db_query
    async def delete(self, **kwargs):
        query = kwargs.get('query', {})

        force = query.get('force', [False])
        if force:
            force = force[0]

        pipeline = kwargs.get('parent_model')
        cache = await BlockCache.initialize_cache()
        if pipeline:
            cache.remove_pipeline(self.model, pipeline.uuid)

        cache_block_action_object = await BlockActionObjectCache.initialize_cache()
        cache_block_action_object.update_block(self.model, remove=True)

        return self.model.delete(force=force)

    @safe_db_query
    async def update(self, payload, **kwargs):
        cache_block_action_object = await BlockActionObjectCache.initialize_cache()
        cache_block_action_object.update_block(self.model, remove=True)

        query = kwargs.get('query', {})
        update_state = query.get('update_state', [False])
        if update_state:
            update_state = update_state[0]
        self.model.update(
            payload,
            update_state=update_state,
        )

        cache_block_action_object.update_block(self.model)

    async def get_pipelines_from_cache(self):
        await BlockCache.initialize_cache()

        return self.model.get_pipelines_from_cache()



import asyncio

from sqlalchemy import or_
from sqlalchemy.orm import aliased

from mage_ai.ai.constants import LLMUseCase
from mage_ai.api.operations.constants import DELETE
from mage_ai.api.resources.BaseResource import BaseResource
from mage_ai.api.resources.BlockResource import BlockResource
from mage_ai.api.resources.LlmResource import LlmResource
from mage_ai.data_preparation.models.block.dbt.utils import (
    add_blocks_upstream_from_refs,
)
from mage_ai.data_preparation.models.constants import (
    BlockLanguage,
    BlockType,
    PipelineStatus,
)
from mage_ai.data_preparation.models.custom_templates.custom_pipeline_template import (
    CustomPipelineTemplate,
)
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.triggers import ScheduleStatus
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.schedules import PipelineRun, PipelineSchedule
from mage_ai.orchestration.pipeline_scheduler import (
    PipelineScheduler,
    retry_pipeline_run,
)
from mage_ai.server.active_kernel import switch_active_kernel
from mage_ai.server.kernels import PIPELINE_TO_KERNEL_NAME
from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.array import find_index
from mage_ai.shared.hash import group_by, ignore_keys, merge_dict
from mage_ai.usage_statistics.logger import UsageStatisticLogger


class PipelineResource(BaseResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        include_schedules = query.get('include_schedules', [False])
        if include_schedules:
            include_schedules = include_schedules[0]

        tags = query.get('tag[]', [])
        if tags:
            new_tags = []
            for tag in tags:
                new_tags += tag.split(',')
            tags = new_tags

        pipeline_types = query.get('type[]', [])
        if pipeline_types:
            pipeline_types = pipeline_types[0]
        if pipeline_types:
            pipeline_types = pipeline_types.split(',')

        pipeline_statuses = query.get('status[]', [])
        if pipeline_statuses:
            pipeline_statuses = pipeline_statuses[0]
        if pipeline_statuses:
            pipeline_statuses = pipeline_statuses.split(',')

        if tags:
            from mage_ai.cache.tag import KEY_FOR_PIPELINES, TagCache

            await TagCache.initialize_cache()

            cache = TagCache()
            tags_mapping = cache.get_tags()
            pipeline_uuids = set()

            for tag_uuid in tags:
                pipelines_dict = tags_mapping.get(tag_uuid, {}).get(KEY_FOR_PIPELINES, {})
                if pipelines_dict:
                    pipeline_uuids.update(pipelines_dict.keys())

            pipeline_uuids = list(pipeline_uuids)
        else:
            pipeline_uuids = Pipeline.get_all_pipelines(get_repo_path())

        await UsageStatisticLogger().pipelines_impression(lambda: len(pipeline_uuids))

        async def get_pipeline(uuid):
            try:
                return await Pipeline.get_async(uuid)
            except Exception as err:
                err_message = f'Error loading pipeline {uuid}: {err}.'
                if err.__class__.__name__ == 'OSError' and 'Too many open files' in err.strerror:
                    raise Exception(err_message)
                else:
                    print(err_message)
                    return None

        pipelines = await asyncio.gather(
            *[get_pipeline(uuid) for uuid in pipeline_uuids]
        )
        pipelines = [p for p in pipelines if p is not None]

        @safe_db_query
        def query_pipeline_schedules(pipeline_uuids):
            a = aliased(PipelineSchedule, name='a')
            result = (
                PipelineSchedule.
                select(*[
                    a.created_at,
                    a.id,
                    a.name,
                    a.pipeline_uuid,
                    a.schedule_interval,
                    a.schedule_type,
                    a.start_time,
                    a.status,
                    a.updated_at,
                ]).
                filter(
                    a.pipeline_uuid.in_(pipeline_uuids),
                    or_(
                        a.repo_path == get_repo_path(),
                        a.repo_path.is_(None),
                    )
                )
            ).all()
            return group_by(lambda x: x.pipeline_uuid, result)

        mapping = {}
        if include_schedules:
            mapping = query_pipeline_schedules(pipeline_uuids)
        if pipeline_types:
            pipelines = [p for p in pipelines if p.type in pipeline_types]

        filtered_pipelines = []
        for pipeline in pipelines:
            schedules = []
            if mapping.get(pipeline.uuid):
                schedules = mapping[pipeline.uuid]
            pipeline.schedules = schedules

            if pipeline_statuses and (
                (PipelineStatus.ACTIVE in pipeline_statuses and
                    any(s.status == ScheduleStatus.ACTIVE
                        for s in pipeline.schedules)) or
                (PipelineStatus.INACTIVE in pipeline_statuses and
                    len(pipeline.schedules) > 0 and
                    all(s.status == ScheduleStatus.INACTIVE
                        for s in pipeline.schedules)) or
                (PipelineStatus.NO_SCHEDULES in pipeline_statuses and
                    len(pipeline.schedules) == 0)
            ):
                filtered_pipelines.append(pipeline)

        if include_schedules and pipeline_statuses:
            pipelines = filtered_pipelines

        return self.build_result_set(
            pipelines,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        clone_pipeline_uuid = payload.get('clone_pipeline_uuid')
        template_uuid = payload.get('custom_template_uuid')
        name = payload.get('name')
        pipeline_type = payload.get('type')

        if template_uuid:
            custom_template = CustomPipelineTemplate.load(template_uuid=template_uuid)
            pipeline = custom_template.create_pipeline(name)
        elif clone_pipeline_uuid is None:
            pipeline = Pipeline.create(
                name,
                pipeline_type=pipeline_type,
                repo_path=get_repo_path(),
            )
        else:
            source = Pipeline.get(clone_pipeline_uuid)
            pipeline = Pipeline.duplicate(source, name)

        return self(pipeline, user, **kwargs)

    @classmethod
    @safe_db_query
    async def get_model(self, pk):
        return await Pipeline.get_async(pk)

    @classmethod
    @safe_db_query
    async def member(self, pk, user, **kwargs):
        pipeline = await Pipeline.get_async(pk)

        if kwargs.get('api_operation_action', None) != DELETE:
            switch_active_kernel(PIPELINE_TO_KERNEL_NAME[pipeline.type])

        query = kwargs.get('query', {})
        include_block_pipelines = query.get('include_block_pipelines', [False])
        if include_block_pipelines:
            include_block_pipelines = include_block_pipelines[0]
        if include_block_pipelines:
            from mage_ai.cache.block import BlockCache

            await BlockCache.initialize_cache()

        return self(pipeline, user, **kwargs)

    @safe_db_query
    def delete(self, **kwargs):
        return self.model.delete()

    @safe_db_query
    async def update(self, payload, **kwargs):
        if 'add_upstream_for_block_uuid' in payload:
            block_uuid = payload['add_upstream_for_block_uuid']
            block = self.model.get_block(block_uuid, widget=False)
            arr = add_blocks_upstream_from_refs(block)
            upstream_block_uuids = [b.uuid for b in arr]

            for b in block.upstream_blocks:
                upstream_block_uuids.append(b.uuid)

            self.model.add_block(
                block,
                upstream_block_uuids,
                priority=len(upstream_block_uuids),
                widget=False,
            )

            return self

        query = kwargs.get('query', {})
        update_content = query.get('update_content', [False])
        if update_content:
            update_content = update_content[0]

        llm_payload = payload.get('llm')
        if llm_payload:
            llm_use_case = llm_payload.get('use_case')
            llm_request = llm_payload.get('request')

            if 'pipeline_uuid' not in llm_payload:
                llm_payload['pipeline_uuid'] = self.model.uuid

            llm_resource = await LlmResource.create(llm_payload, self.current_user, **kwargs)
            llm_response = llm_resource.model.get('response')

            pipeline_doc = None
            block_docs = []
            blocks = self.model.blocks_by_uuid.values()

            async def _add_markdown_block(block_doc: str, block_uuid: str, priority: int):
                return await BlockResource.create(dict(
                    content=block_doc.strip() if block_doc else block_doc,
                    language=BlockLanguage.MARKDOWN,
                    name=f'Documentation for {block_uuid}',
                    priority=priority,
                    type=BlockType.MARKDOWN,
                ), self.current_user, **merge_dict(kwargs, dict(
                    parent_model=self.model,
                )))

            if LLMUseCase.GENERATE_COMMENT_FOR_BLOCK == llm_use_case:
                pass
            elif LLMUseCase.GENERATE_DOC_FOR_BLOCK == llm_use_case:
                block_doc = llm_response.get('block_doc')
                if block_doc:
                    block_uuid = llm_request.get('block_uuid')
                    priority = find_index(lambda x: x.uuid == block_uuid, blocks)
                    await _add_markdown_block(block_doc, block_uuid, priority)
            elif LLMUseCase.GENERATE_DOC_FOR_PIPELINE == llm_use_case:
                block_docs = llm_response.get('block_docs', [])
                pipeline_doc = llm_response.get('pipeline_doc')

                if pipeline_doc:
                    lines = []
                    if payload.get('description'):
                        lines.append(payload.get('description'))
                    lines.append(pipeline_doc)
                    payload['description'] = '\n'.join(lines).strip()

            if block_docs and len(block_docs) >= 1:
                blocks_with_docs = list(zip(block_docs, blocks))
                blocks_with_docs.reverse()
                blocks_with_docs_length = len(blocks_with_docs)

                for idx, tup in enumerate(blocks_with_docs):
                    priority = blocks_with_docs_length - (idx + 1)
                    block_doc, block = tup

                    if block_doc:
                        await _add_markdown_block(block_doc, block.uuid, priority)

            if pipeline_doc:
                await _add_markdown_block(pipeline_doc, self.model.uuid, 0)

        await self.model.update(
            ignore_keys(payload, ['add_upstream_for_block_uuid']),
            update_content=update_content,
        )
        switch_active_kernel(PIPELINE_TO_KERNEL_NAME[self.model.type])

        @safe_db_query
        def update_schedule_status(status, pipeline_uuid):
            schedules = (
                PipelineSchedule.
                query.
                filter(PipelineSchedule.pipeline_uuid == pipeline_uuid)
            ).all()
            for schedule in schedules:
                schedule.update(status=status)

        @safe_db_query
        def cancel_pipeline_runs(status, pipeline_uuid):
            pipeline_runs = (
                PipelineRun.
                query.
                filter(PipelineRun.pipeline_uuid == pipeline_uuid).
                filter(PipelineRun.status.in_([
                    PipelineRun.PipelineRunStatus.INITIAL,
                    PipelineRun.PipelineRunStatus.RUNNING,
                ]))
            )
            for pipeline_run in pipeline_runs:
                PipelineScheduler(pipeline_run).stop()

        def retry_pipeline_runs(pipeline_runs):
            for run in pipeline_runs:
                retry_pipeline_run(run)

        status = payload.get('status')
        pipeline_uuid = self.model.uuid

        def _update_callback(resource):
            if status:
                if status in [
                    ScheduleStatus.ACTIVE.value,
                    ScheduleStatus.INACTIVE.value,
                ]:
                    update_schedule_status(status, pipeline_uuid)
                elif status == PipelineRun.PipelineRunStatus.CANCELLED.value:
                    cancel_pipeline_runs(status, pipeline_uuid)
                elif status == 'retry' and payload.get('pipeline_runs'):
                    retry_pipeline_runs(payload.get('pipeline_runs'))

        self.on_update_callback = _update_callback

        return self



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.PullRequestPresenter import PullRequestPresenter


class PullRequestPolicy(BasePolicy):
    pass


PullRequestPolicy.allow_actions([
    constants.DETAIL,
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())


PullRequestPolicy.allow_actions([
    constants.DETAIL,
    constants.LIST,
    constants.CREATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_editor_role())


PullRequestPolicy.allow_read(PullRequestPresenter.default_attributes + [
], scopes=[
    OauthScope.CLIENT_PRIVATE
], on_action=[
    constants.DETAIL,
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())


PullRequestPolicy.allow_read(PullRequestPresenter.default_attributes + [
], scopes=[
    OauthScope.CLIENT_PRIVATE
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role())


PullRequestPolicy.allow_write([
    'base_branch',
    'body',
    'compare_branch',
    'repository',
    'title',
], scopes=[
    OauthScope.CLIENT_PRIVATE
], on_action=[
    constants.CREATE,
], condition=lambda policy: policy.has_at_least_editor_role())


PullRequestPolicy.allow_query([
    'repository',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.LIST,
], condition=lambda policy: policy.has_at_least_viewer_role())



import os
import unittest
from unittest.mock import MagicMock, patch

from mage_ai.data_preparation.executors.k8s_block_executor import K8sBlockExecutor
from mage_ai.services.k8s.config import K8sExecutorConfig


class K8sBlockExecutorTestCase(unittest.TestCase):
    def setUp(self):
        self.pipeline = MagicMock()
        self.pipeline.uuid = 'pipeline_uuid'
        self.pipeline.repo_config = MagicMock()
        self.pipeline.repo_config.variables_dir = os.path.join(os.getcwd(), 'mage_data')
        self.block_uuid = 'block-uuid'

        self.logger_manager = MagicMock()
        self.logger = MagicMock()
        self.logger_manager.logger = self.logger

        self.executor = K8sBlockExecutor(self.pipeline, self.block_uuid)

        self.executor.logger_manager = self.logger_manager
        self.executor.logger = self.logger

    @patch('mage_ai.data_preparation.executors.k8s_block_executor.K8sJobManager')
    def test_execute(self, job_manager_mock):
        # Mock the necessary objects and methods
        self.executor._run_commands = MagicMock(return_value='mocked_cmd')
        job_manager_instance_mock = MagicMock()
        job_manager_mock.return_value = job_manager_instance_mock

        # Call the method to test
        self.executor._execute(block_run_id=1, global_vars={'key': 'value'})

        # Assertions
        self.executor._run_commands.assert_called_once_with(
            block_run_id=1,
            global_vars={'key': 'value'},
        )
        job_manager_mock.assert_called_once_with(
            job_name='mage-data-prep-block-1',
            logger=self.executor.logger,
            logging_tags={},
        )
        job_manager_instance_mock.run_job.assert_called_once_with(
            'mocked_cmd',
            k8s_config=K8sExecutorConfig.load(config={}),
        )

    @patch('mage_ai.data_preparation.executors.k8s_block_executor.K8sJobManager')
    def test_execute_with_custom_job_name_prefix(self, job_manager_mock):
        # Mock the necessary objects and methods
        self.executor._run_commands = MagicMock(return_value='mocked_cmd')
        job_manager_instance_mock = MagicMock()
        job_manager_mock.return_value = job_manager_instance_mock

        # Set a custom job name prefix
        self.executor.executor_config.job_name_prefix = 'custom-prefix'

        # Call the method to test
        self.executor._execute(block_run_id=1, global_vars={'key': 'value'})

        # Assertions
        self.executor._run_commands.assert_called_once_with(
            block_run_id=1,
            global_vars={'key': 'value'},
        )
        job_manager_mock.assert_called_once_with(
            job_name='mage-custom-prefix-block-1',
            logger=self.executor.logger,
            logging_tags={},
        )
        job_manager_instance_mock.run_job.assert_called_once_with(
            'mocked_cmd',
            k8s_config=K8sExecutorConfig.load(config={'job_name_prefix': 'custom-prefix'}),
        )



import os
from typing import Dict, List

MIN_SECRET_LENGTH = 8
WHITELISTED_ENV_VARS = set([
    'MAGE_DATA_DIR',
    'MAGE_REPO_PATH',
    'HOME',
    'PWD',
    'PYTHONPATH',
])


def filter_out_env_var_values(value: str) -> str:
    env_var_values = dict(os.environ).values()
    whitelisted_env_var_values = set([os.getenv(k) for k in WHITELISTED_ENV_VARS if os.getenv(k)])
    env_var_values = [v for v in env_var_values
                      if v and len(v) >= MIN_SECRET_LENGTH
                      and v not in whitelisted_env_var_values]
    return filter_out_values(value, env_var_values)


def filter_out_config_values(log: str, config: Dict) -> str:
    """
    Used for integration pipeline logging security. Filters out values from
    the passed in config if the config value is a string and is over
    MIN_SECRET_LENGTH characters long.

    Args:
        log (str): The log to filter.
        config (Dict[str, any]): The config to get filter values from.

    Returns:
        str: Log with config values filtered out.
    """
    if not log or not config:
        return log
    values = config.values()
    secret_values = [v for v in values if v and isinstance(v, str) and len(v) >= MIN_SECRET_LENGTH]
    return filter_out_values(log, secret_values)


def filter_out_values(log: str, values: List[str]) -> str:
    if not log or not values:
        return log
    values.sort(key=len, reverse=True)
    log_clean = log
    for value in values:
        replace_value = '*' * len(value)
        log_clean = log_clean.replace(value, replace_value)
    return log_clean



import asyncio
import io
import logging
import logging.handlers
import os
from datetime import datetime
from typing import Callable, Dict, List

from mage_ai.data_preparation.logging import LoggingConfig
from mage_ai.data_preparation.models.constants import LOGS_DIR
from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.repo_manager import (
    RepoConfig,
    get_repo_config,
    get_repo_path,
)
from mage_ai.data_preparation.storage.local_storage import LocalStorage
from mage_ai.shared.array import find
from mage_ai.shared.dates import str_to_timedelta

MAX_LOG_FILE_SIZE = 20 * 1024 * 1024


class LoggerManager:
    def __init__(
        self,
        repo_path: str = None,
        logs_dir: str = None,
        pipeline_uuid: str = None,
        block_uuid: str = None,
        partition: str = None,
        repo_config: RepoConfig = None,
        subpartition: str = None,
    ):
        """
        Initialize the LoggerManager with provided configuration.

        Args:
            repo_path (str, optional): Root directory path of the repository.
            logs_dir (str, optional): Directory path where logs will be stored.
            pipeline_uuid (str, optional): Unique identifier for the pipeline.
            block_uuid (str, optional): Unique identifier for the block.
            partition (str, optional): Partition identifier.
            repo_config (RepoConfig, optional): Repository configuration.
            subpartition (str, optional): Subpartition identifier.
        """
        self.repo_path = repo_path or get_repo_path()
        self.logs_dir = logs_dir
        self.pipeline_uuid = pipeline_uuid
        self.block_uuid = block_uuid
        self.partition = partition
        self.subpartition = subpartition

        # Load the repository configuration for logging
        self.repo_config = repo_config or get_repo_config()
        logging_config = self.repo_config.logging_config if self.repo_config else dict()
        self.logging_config = LoggingConfig.load(config=logging_config)

        # Create logger instance based on provided identifiers
        if self.pipeline_uuid:
            logger_name_parts = [self.pipeline_uuid]
        else:
            logger_name_parts = ['all_pipelines']
        if self.partition is not None:
            logger_name_parts.append(self.partition)
        if self.block_uuid is not None:
            logger_name_parts.append(self.block_uuid)
        logger_name = '/'.join(logger_name_parts)
        self.logger = logging.getLogger(logger_name)

        # Set the log level based on the configuration
        self.log_level = logging.getLevelName(self.logging_config.level)
        self.logger.setLevel(self.log_level)

        # Create log formatter
        self.formatter = logging.Formatter(
            '%(asctime)s %(message)s',
            '%Y-%m-%dT%H:%M:%S',
        )

        # Initialize log stream and add appropriate handlers to the logger
        self.stream = None
        if not self.logger.handlers:
            if self.logging_config.destination_config:
                # If there is a destination configuration, use a stream handler
                handler = self.create_stream_handler()
            else:
                # If no destination configuration, use a rotating file handler
                log_filepath = self.get_log_filepath(create_dir=True)
                handler = logging.handlers.RotatingFileHandler(
                    log_filepath,
                    backupCount=10,
                    maxBytes=MAX_LOG_FILE_SIZE,
                )

            handler.setLevel(self.log_level)
            handler.setFormatter(self.formatter)
            self.logger.addHandler(handler)
        else:
            if self.logging_config.destination_config:
                # If a stream handler already exists, get the stream object
                stream_handler = find(
                    lambda hr: hr.__class__ == logging.StreamHandler, self.logger.handlers
                )
                if stream_handler:
                    self.stream = stream_handler.stream

        self.storage = LocalStorage()

    def create_stream_handler(self):
        """
        Create a stream handler to output logs to a stream (e.g., in-memory buffer).
        Used when a destination configuration is present.

        Returns:
            logging.StreamHandler: The stream handler.
        """
        self.stream = io.StringIO()
        return logging.StreamHandler(self.stream)

    def create_log_filepath_dir(self, path):
        """
        Create the directory path for log files, if it doesn't exist.

        Args:
            path (str): The path to create the directory.
        """
        if not os.path.exists(path):
            os.makedirs(path)

    def delete_old_logs(self):
        """
        Delete old log files based on log retention_period
        """
        log_retention_period = self.logging_config.retention_period
        if not log_retention_period:
            return
        min_partition = (datetime.utcnow() -
                         str_to_timedelta(log_retention_period)).strftime(
                            format='%Y%m%dT%H%M%S')

        if self.pipeline_uuid is None:
            from mage_ai.data_preparation.models.pipeline import Pipeline

            pipeline_uuids = Pipeline.get_all_pipelines(self.repo_path)
        else:
            pipeline_uuids = [self.pipeline_uuid]

        for pipeline_uuid in pipeline_uuids:
            print(f'Removing old logs from pipeline {pipeline_uuid}')
            pipeline_log_path = self.get_log_filepath_prefix(pipeline_uuid=pipeline_uuid)
            dirs = self.storage.listdir(pipeline_log_path)
            for dirname in dirs:
                if dirname.isdigit():
                    pipeline_schedule_vpath = os.path.join(pipeline_log_path, dirname)
                    execution_partitions = self.storage.listdir(
                        pipeline_schedule_vpath,
                    )
                    for partition in execution_partitions:
                        if partition <= min_partition:
                            pipeline_partition_vpath = os.path.join(
                                pipeline_schedule_vpath,
                                partition,
                            )
                            print(f'Removing folder {pipeline_partition_vpath}')
                            self.storage.remove_dir(pipeline_partition_vpath)

    def get_log_filepath_prefix(
        self,
        pipeline_uuid: str = None,
    ):
        """
        Get the prefix of the log file path.

        Returns:
            str: The log file path prefix based on pipeline_uuid, logs_dir, partition, and
                 subpartition.
        """
        logs_dir = self.logs_dir or self.repo_config.variables_dir

        return os.path.join(
            logs_dir,
            'pipelines',
            pipeline_uuid or self.pipeline_uuid or 'all_pipelines',
            LOGS_DIR,
            self.partition or '',
            self.subpartition or '',
        )

    def get_log_filepath(self, create_dir: bool = False):
        """
        Get the full log file path for the current pipeline or block.

        Args:
            create_dir (bool, optional): If True, create the log file directory if it doesn't exist.

        Returns:
            str: The full log file path.
        Raises:
            Exception: If pipeline_uuid is None.
        """
        prefix = self.get_log_filepath_prefix()

        if create_dir:
            self.create_log_filepath_dir(prefix)

        if self.block_uuid is None:
            log_filepath = os.path.join(prefix, 'pipeline.log')
        else:
            log_filepath = os.path.join(prefix, f'{self.block_uuid}.log')
        return log_filepath

    def get_logs(self):
        """
        Get logs from the current log file.

        Returns:
            dict: A dictionary containing the logs, including their content.
        """
        file = File.from_path(self.get_log_filepath())
        return file.to_dict(include_content=True)

    async def get_logs_async(self):
        """
        Get logs asynchronously from the current log file.

        Returns:
            dict: A dictionary containing the logs, including their content.
        """
        file = File.from_path(self.get_log_filepath())
        return await file.to_dict_async(include_content=True)

    async def get_logs_in_subpartition_async(self, filter_func: Callable = None) -> List[Dict]:
        """
        Get logs asynchronously from multiple log files within the subpartition.

        Args:
            filter_func (Callable, optional): A filter function to apply on each log file.

        Returns:
            List[Dict]: A list of dictionaries containing the logs, including their content.
        """
        files = []

        base_path = self.get_log_filepath_prefix()
        if os.path.exists(base_path):
            for filename in os.listdir(base_path):
                full_path = f'{base_path}/{filename}'
                if not os.path.isfile(full_path):
                    continue

                file = File.from_path(full_path)
                should_add = True

                if filter_func:
                    should_add = filter_func(file)

                if should_add:
                    files.append(file)

        return await asyncio.gather(
            *[file.to_dict_async(include_content=True) for file in files]
        )

    def output_logs_to_destination(self):
        """
        Output logs to the configured destination.
        (Placeholder, this function is not implemented yet.)
        """
        pass



import urllib.parse

from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.cache.block_action_object import BlockActionObjectCache
from mage_ai.data_preparation.models.custom_templates.constants import (
    DIRECTORY_FOR_BLOCK_TEMPLATES,
    DIRECTORY_FOR_PIPELINE_TEMPLATES,
)
from mage_ai.data_preparation.models.custom_templates.custom_block_template import (
    CustomBlockTemplate,
)
from mage_ai.data_preparation.models.custom_templates.custom_pipeline_template import (
    CustomPipelineTemplate,
)
from mage_ai.data_preparation.models.custom_templates.utils import (
    flatten_files,
    get_templates,
    group_and_hydrate_files,
)
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.templates.template import fetch_template_source
from mage_ai.shared.hash import ignore_keys
from mage_ai.shared.utils import clean_name

OBJECT_TYPE_KEY = 'object_type'


class CustomTemplateResource(GenericResource):
    @classmethod
    def collection(self, query, meta, user, **kwargs):
        object_type = query.get(OBJECT_TYPE_KEY, [None])
        if object_type:
            object_type = object_type[0]

        templates = []
        file_dicts = []

        if DIRECTORY_FOR_BLOCK_TEMPLATES == object_type:
            file_dicts = get_templates(DIRECTORY_FOR_BLOCK_TEMPLATES)
            template_class = CustomBlockTemplate
        elif DIRECTORY_FOR_PIPELINE_TEMPLATES == object_type:
            file_dicts = get_templates(DIRECTORY_FOR_PIPELINE_TEMPLATES)
            template_class = CustomPipelineTemplate

        if file_dicts:
            file_dicts_flat = flatten_files(file_dicts)
            templates = group_and_hydrate_files(file_dicts_flat, template_class)

        return self.build_result_set(
            templates,
            user,
            **kwargs,
        )

    @classmethod
    async def create(self, payload, user, **kwargs):
        custom_template = None
        object_type = payload.get(OBJECT_TYPE_KEY)
        template_uuid = payload.get('template_uuid')

        if template_uuid:
            template_uuid = clean_name(template_uuid)
            payload['template_uuid'] = template_uuid

        if DIRECTORY_FOR_BLOCK_TEMPLATES == object_type:
            custom_template = CustomBlockTemplate.load(template_uuid=template_uuid)

            if not custom_template:
                custom_template = CustomBlockTemplate(**ignore_keys(payload, [
                    'uuid',
                    OBJECT_TYPE_KEY,
                ]))
                if user:
                    custom_template.user = dict(
                        username=user.username,
                    )

                custom_template.content = fetch_template_source(
                    custom_template.block_type,
                    payload.get('config', {}),
                    language=custom_template.language,
                )

                custom_template.save()

                cache = await BlockActionObjectCache.initialize_cache()
                cache.update_custom_block_template(custom_template)
        elif DIRECTORY_FOR_PIPELINE_TEMPLATES == object_type:
            custom_template = CustomPipelineTemplate.load(template_uuid=template_uuid)

            if not custom_template:
                pipeline = Pipeline.get(payload.get('pipeline_uuid'))
                custom_template = CustomPipelineTemplate.create_from_pipeline(
                    pipeline,
                    template_uuid,
                    name=payload.get('name'),
                    description=payload.get('description'),
                )

                if user:
                    custom_template.user = dict(
                        username=user.username,
                    )

                custom_template.save()

        if custom_template:
            return self(custom_template, user, **kwargs)

    @classmethod
    def member(self, pk, user, **kwargs):
        query = kwargs.get('query', {})
        object_type = query.get(OBJECT_TYPE_KEY, [None])
        if object_type:
            object_type = object_type[0]

        template_uuid = urllib.parse.unquote(pk)

        try:
            if DIRECTORY_FOR_BLOCK_TEMPLATES == object_type:
                return self(CustomBlockTemplate.load(template_uuid=template_uuid), user, **kwargs)
            elif DIRECTORY_FOR_PIPELINE_TEMPLATES == object_type:
                return self(
                    CustomPipelineTemplate.load(template_uuid=template_uuid),
                    user,
                    **kwargs,
                )
        except Exception as err:
            print(f'[WARNING] CustomTemplateResource.member: {err}')
            raise ApiError(ApiError.RESOURCE_NOT_FOUND)

    async def delete(self, **kwargs):
        cache = await BlockActionObjectCache.initialize_cache()
        cache.update_custom_block_template(self.model, remove=True)
        self.model.delete

    async def update(self, payload, **kwargs):
        template_uuid = payload.get('template_uuid')

        if template_uuid:
            template_uuid = clean_name(template_uuid)
            payload['template_uuid'] = template_uuid

        object_type = payload.get('object_type')

        cache = None
        if DIRECTORY_FOR_BLOCK_TEMPLATES == object_type:
            cache = await BlockActionObjectCache.initialize_cache()
            cache.update_custom_block_template(self.model, remove=True)

        for key, value in ignore_keys(payload, [
            'uuid',
            OBJECT_TYPE_KEY,
        ]).items():
            setattr(self.model, key, value)
        self.model.save()

        if DIRECTORY_FOR_BLOCK_TEMPLATES == object_type and cache:
            cache.update_custom_block_template(self.model)



import enum
import os
import traceback
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List

import yaml

from mage_ai.data_preparation.models.constants import PIPELINES_FOLDER
from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.config import BaseConfig
from mage_ai.shared.constants import VALID_ENVS
from mage_ai.shared.hash import index_by
from mage_ai.shared.io import safe_write

TRIGGER_FILE_NAME = 'triggers.yaml'


class ScheduleStatus(str, enum.Enum):
    ACTIVE = 'active'
    INACTIVE = 'inactive'


class ScheduleType(str, enum.Enum):
    API = 'api'
    EVENT = 'event'
    TIME = 'time'


class ScheduleInterval(str, enum.Enum):
    ONCE = '@once'
    HOURLY = '@hourly'
    DAILY = '@daily'
    WEEKLY = '@weekly'
    MONTHLY = '@monthly'


@dataclass
class SettingsConfig(BaseConfig):
    skip_if_previous_running: bool = False
    allow_blocks_to_fail: bool = False
    landing_time_enabled: bool = False


@dataclass
class Trigger(BaseConfig):
    name: str
    pipeline_uuid: str
    schedule_type: ScheduleType
    start_time: datetime
    schedule_interval: str
    status: ScheduleStatus = ScheduleStatus.INACTIVE
    variables: Dict = field(default_factory=dict)
    sla: int = None     # in seconds
    settings: Dict = field(default_factory=dict)
    envs: List = field(default_factory=list)

    def __post_init__(self):
        if self.schedule_type and type(self.schedule_type) is str:
            self.schedule_type = ScheduleType(self.schedule_type)
        if self.status and type(self.status) is str:
            self.status = ScheduleStatus(self.status)
        if any(env not in VALID_ENVS for env in self.envs):
            raise Exception(f'Please provide valid env values inside {list(VALID_ENVS)}.')

    def to_dict(self) -> Dict:
        return dict(
            envs=self.envs,
            name=self.name,
            pipeline_uuid=self.pipeline_uuid,
            schedule_interval=self.schedule_interval,
            schedule_type=self.schedule_type.value if self.schedule_type else self.schedule_type,
            settings=self.settings,
            sla=self.sla,
            start_time=self.start_time,
            status=self.status.value if self.status else self.status,
            variables=self.variables,
        )


def get_triggers_file_path(pipeline_uuid: str) -> str:
    pipeline_path = os.path.join(get_repo_path(), PIPELINES_FOLDER, pipeline_uuid)
    trigger_file_path = os.path.join(pipeline_path, TRIGGER_FILE_NAME)
    return trigger_file_path


def load_triggers_file_content(pipeline_uuid: str) -> str:
    content = None

    trigger_file_path = get_triggers_file_path(pipeline_uuid)
    if os.path.exists(trigger_file_path):
        with open(trigger_file_path) as fp:
            content = fp.read()

    return content


def load_triggers_file_data(pipeline_uuid: str) -> Dict:
    data = {}

    content = load_triggers_file_content(pipeline_uuid)
    if content:
        data = yaml.safe_load(content) or {}

    return data


def get_triggers_by_pipeline(pipeline_uuid: str) -> List[Trigger]:
    trigger_file_path = get_triggers_file_path(pipeline_uuid)

    if not os.path.exists(trigger_file_path):
        return []

    try:
        content = load_triggers_file_content(pipeline_uuid)
        triggers = load_trigger_configs(content, pipeline_uuid=pipeline_uuid)
    except Exception:
        traceback.print_exc()
        triggers = []

    return triggers


def build_triggers(
    trigger_configs: Dict,
    pipeline_uuid: str = None,
    raise_exception: bool = False,
) -> List[Trigger]:
    triggers = []
    for trigger_config in trigger_configs:
        if pipeline_uuid:
            trigger_config['pipeline_uuid'] = pipeline_uuid
        try:
            trigger = Trigger.load(config=trigger_config)
            triggers.append(trigger)
        except Exception as e:
            if raise_exception:
                raise Exception(
                    f'Failed to parse trigger config {trigger_config}. {str(e)}') from e
            else:
                print(f'Failed to parse trigger config for pipeline {pipeline_uuid}')
                traceback.print_exc()
    return triggers


def load_trigger_configs(
    content: str,
    pipeline_uuid: str = None,
    raise_exception: bool = False,
) -> List[Trigger]:
    yaml_config = yaml.safe_load(content) or {}
    trigger_configs = yaml_config.get('triggers') or {}

    return build_triggers(trigger_configs, pipeline_uuid, raise_exception)


def add_or_update_trigger_for_pipeline_and_persist(
    trigger: Trigger,
    pipeline_uuid: str,
) -> Dict:
    yaml_config = load_triggers_file_data(pipeline_uuid)
    trigger_configs = yaml_config.get('triggers') or {}

    triggers_by_name = index_by(
        lambda trigger: trigger.name,
        build_triggers(trigger_configs, pipeline_uuid),
    )
    triggers_by_name[trigger.name] = trigger

    yaml_config['triggers'] = [trigger.to_dict() for trigger in triggers_by_name.values()]

    content = yaml.safe_dump(yaml_config)
    trigger_file_path = get_triggers_file_path(pipeline_uuid)
    safe_write(trigger_file_path, content)

    return triggers_by_name



import json
import os
import sys
from typing import List, Union

import typer
from click import Context
from rich import print
from typer.core import TyperGroup

from mage_ai.cli.utils import parse_runtime_variables
from mage_ai.data_preparation.repo_manager import ProjectType
from mage_ai.services.newrelic import initialize_new_relic
from mage_ai.shared.constants import InstanceType


class OrderCommands(TyperGroup):
    def list_commands(self, ctx: Context):
        """Return list of commands in the order they appear."""
        return list(self.commands)


app = typer.Typer(
    cls=OrderCommands,
)

# Defaults

INIT_PROJECT_PATH_DEFAULT = typer.Argument(..., help='path of the Mage project to be created.')
INIT_PROJECT_TYPE_DEFAULT = typer.Option(
    ProjectType.STANDALONE.value,
    help='type of project to create, options are main, sub, or standalone'
)
INIT_CLUSTER_TYPE_DEFAULT = typer.Option(
    None,
    help='type of instance to create for workspace management',
)
INIT_PROJECT_UUID_DEFAULT = typer.Option(
    None,
    help='project uuid for the new project',
)

START_PROJECT_PATH_DEFAULT = typer.Argument(
    os.getcwd(), help='path of the Mage project to be loaded.')
START_HOST_DEFAULT = typer.Option('localhost', help='specify the host.')
START_PORT_DEFAULT = typer.Option('6789', help='specify the port.')
START_MANAGE_INSTANCE_DEFAULT = typer.Option('0', help='')
START_DBT_DOCS_INSTANCE_DEFAULT = typer.Option('0', help='')
START_INSTANCE_TYPE_DEFAULT = typer.Option(
    InstanceType.SERVER_AND_SCHEDULER.value, help='specify the instance type.')
START_PROJECT_TYPE_DEFAULT = typer.Option(
    ProjectType.STANDALONE.value,
    help='create project of this type if does not exist, options are main, sub, or standalone',
)
START_CLUSTER_TYPE_DEFAULT = typer.Option(
    None,
    help='type of instance to create for workspace management',
)
START_PROJECT_UUID_DEFAULT = typer.Option(
    None,
    help='set project uuid if it has not been set for the project already',
)

RUN_PROJECT_PATH_DEFAULT = typer.Argument(
    ..., help='path of the Mage project that contains the pipeline.'
)
RUN_PIPELINE_UUID_DEFAULT = typer.Argument(
    ..., help='uuid of the pipeline to be run.'
)
RUN_TEST_DEFAULT = typer.Option(
    False, help='specify if tests should be run.'
)
RUN_BLOCK_UUID_DEFAULT = typer.Option(
    None, help='uuid of the block to be run.'
)
RUN_EXECUTION_PARTITION_DEFAULT = typer.Option(
    None, help=''
)
RUN_EXECUTOR_TYPE_DEFAULT = typer.Option(
    None, help=''
)
RUN_CALLBACK_URL_DEFAULT = typer.Option(
    None, help=''
)
RUN_BLOCK_RUN_ID_DEFAULT = typer.Option(
    None, help=''
)
RUN_PIPELINE_RUN_ID_DEFAULT = typer.Option(
    None, help=''
)
RUN_RUNTIME_VARS_DEFAULT = typer.Option(
    None, help='specify runtime variables. These will overwrite the pipeline global variables.'
)
RUN_SKIP_SENSORS_DEFAULT = typer.Option(
    False, help='specify if the sensors should be skipped.'
)
RUN_TEMPLATE_RUNTIME_CONFIGURATION_DEFAULT = typer.Option(
    None, help='runtime configuration of data integration block runs.'
)
CLEAN_LOGS_PROJECT_PATH_DEFAULT = typer.Argument(
    ..., help='path of the Mage project to clean old logs.'
)
CLEAN_LOGS_PIPELINE_UUID_DEFAULT = typer.Option(
    None, help='uuid of the pipeline to clean.'
)
CLEAN_VARIABLES_PROJECT_PATH_DEFAULT = typer.Argument(
    ..., help='path of the Mage project to clean variables.'
)
CLEAN_VARIABLES_PIPELINE_UUID_DEFAULT = typer.Option(
    None, help='uuid of the pipeline to clean.'
)

CREATE_SPARK_CLUSTER_PROJECT_PATH_DEFAULT = typer.Argument(
    ..., help='path of the Mage project that contains the EMR config.'
)


@app.command()
def init(
    project_path: str = INIT_PROJECT_PATH_DEFAULT,
    project_type: Union[str, None] = INIT_PROJECT_TYPE_DEFAULT,
    cluster_type: str = INIT_CLUSTER_TYPE_DEFAULT,
    project_uuid: str = INIT_PROJECT_UUID_DEFAULT,
):
    """
    Initialize Mage project.
    """
    from mage_ai.data_preparation.repo_manager import init_repo

    repo_path = os.path.join(os.getcwd(), project_path)
    init_repo(
        repo_path,
        project_type=project_type,
        cluster_type=cluster_type,
        project_uuid=project_uuid,
    )
    print(f'Initialized Mage project at {repo_path}')


@app.command()
def start(
    project_path: str = START_PROJECT_PATH_DEFAULT,
    host: str = START_HOST_DEFAULT,
    port: str = START_PORT_DEFAULT,
    manage_instance: str = START_MANAGE_INSTANCE_DEFAULT,
    dbt_docs_instance: str = START_DBT_DOCS_INSTANCE_DEFAULT,
    instance_type: str = START_INSTANCE_TYPE_DEFAULT,
    project_type: str = START_PROJECT_TYPE_DEFAULT,
    cluster_type: str = START_CLUSTER_TYPE_DEFAULT,
    project_uuid: str = START_PROJECT_UUID_DEFAULT,
):
    """
    Start Mage server and UI.
    """
    from mage_ai.settings.repo import set_repo_path

    # Set repo_path before intializing the DB so that we can get correct db_connection_url
    project_path = os.path.abspath(project_path)
    set_repo_path(project_path)

    from mage_ai.server.server import start_server

    start_server(
        host=host,
        port=port,
        project=project_path,
        manage=manage_instance == "1",
        dbt_docs=dbt_docs_instance == "1",
        instance_type=instance_type,
        project_type=project_type,
        cluster_type=cluster_type,
        project_uuid=project_uuid,
    )


@app.command()
def run(
    project_path: str = RUN_PROJECT_PATH_DEFAULT,
    pipeline_uuid: str = RUN_PIPELINE_UUID_DEFAULT,
    test: bool = RUN_TEST_DEFAULT,
    block_uuid: Union[str, None] = RUN_BLOCK_UUID_DEFAULT,
    execution_partition: Union[str, None] = RUN_EXECUTION_PARTITION_DEFAULT,
    executor_type: Union[str, None] = RUN_EXECUTOR_TYPE_DEFAULT,
    callback_url: Union[str, None] = RUN_CALLBACK_URL_DEFAULT,
    block_run_id: Union[int, None] = RUN_BLOCK_RUN_ID_DEFAULT,
    pipeline_run_id: Union[int, None] = RUN_PIPELINE_RUN_ID_DEFAULT,
    runtime_vars: Union[List[str], None] = RUN_RUNTIME_VARS_DEFAULT,
    skip_sensors: bool = RUN_SKIP_SENSORS_DEFAULT,
    template_runtime_configuration: Union[str, None] = RUN_TEMPLATE_RUNTIME_CONFIGURATION_DEFAULT,
):
    """
    Run pipeline.
    """
    from mage_ai.settings.repo import set_repo_path

    # Set repo_path before intializing the DB so that we can get correct db_connection_url
    project_path = os.path.abspath(project_path)
    set_repo_path(project_path)

    from contextlib import nullcontext

    import newrelic.agent
    import sentry_sdk

    from mage_ai.data_preparation.executors.executor_factory import ExecutorFactory
    from mage_ai.data_preparation.models.pipeline import Pipeline
    from mage_ai.data_preparation.variable_manager import get_global_variables
    from mage_ai.orchestration.db import db_connection
    from mage_ai.orchestration.db.models.schedules import PipelineRun
    from mage_ai.settings import SENTRY_DSN, SENTRY_TRACES_SAMPLE_RATE
    from mage_ai.shared.hash import merge_dict

    sentry_dsn = SENTRY_DSN
    if sentry_dsn:
        sentry_sdk.init(
            sentry_dsn,
            traces_sample_rate=SENTRY_TRACES_SAMPLE_RATE,
        )
    (enable_new_relic, application) = initialize_new_relic()

    with newrelic.agent.BackgroundTask(application, name="mage-run", group='Task') \
         if enable_new_relic else nullcontext():
        runtime_variables = dict()
        if runtime_vars is not None:
            runtime_variables = parse_runtime_variables(runtime_vars)

        sys.path.append(os.path.dirname(project_path))
        pipeline = Pipeline.get(pipeline_uuid, repo_path=project_path)

        db_connection.start_session()

        if pipeline_run_id is None:
            default_variables = get_global_variables(pipeline_uuid)
            global_vars = merge_dict(default_variables, runtime_variables)
        else:
            pipeline_run = PipelineRun.query.get(pipeline_run_id)
            global_vars = pipeline_run.get_variables(extra_variables=runtime_variables)

        if template_runtime_configuration is not None:
            template_runtime_configuration = json.loads(template_runtime_configuration)

        if block_uuid is None:
            ExecutorFactory.get_pipeline_executor(
                pipeline,
                execution_partition=execution_partition,
                executor_type=executor_type,
            ).execute(
                analyze_outputs=False,
                global_vars=global_vars,
                pipeline_run_id=pipeline_run_id,
                run_sensors=not skip_sensors,
                run_tests=test,
                update_status=False,
            )
        else:
            ExecutorFactory.get_block_executor(
                pipeline,
                block_uuid,
                execution_partition=execution_partition,
                executor_type=executor_type,
            ).execute(
                analyze_outputs=False,
                block_run_id=block_run_id,
                callback_url=callback_url,
                global_vars=global_vars,
                pipeline_run_id=pipeline_run_id,
                template_runtime_configuration=template_runtime_configuration,
                update_status=False,
            )
        print('Pipeline run completed.')


@app.command()
def clean_cached_variables(
    project_path: str = CLEAN_VARIABLES_PROJECT_PATH_DEFAULT,
    pipeline_uuid: str = CLEAN_VARIABLES_PIPELINE_UUID_DEFAULT,
):
    from mage_ai.settings.repo import set_repo_path

    project_path = os.path.abspath(project_path)
    set_repo_path(project_path)

    from mage_ai.data_preparation.variable_manager import clean_variables

    clean_variables(pipeline_uuid=pipeline_uuid)


@app.command()
def clean_old_logs(
    project_path: str = CLEAN_LOGS_PROJECT_PATH_DEFAULT,
    pipeline_uuid: str = CLEAN_LOGS_PIPELINE_UUID_DEFAULT,
):
    from mage_ai.settings.repo import set_repo_path

    project_path = os.path.abspath(project_path)
    set_repo_path(project_path)

    from mage_ai.data_preparation.logging.logger_manager_factory import (
        LoggerManagerFactory,
    )

    LoggerManagerFactory.get_logger_manager(
        pipeline_uuid=pipeline_uuid,
    ).delete_old_logs()


@app.command()
def create_spark_cluster(
    project_path: str = CREATE_SPARK_CLUSTER_PROJECT_PATH_DEFAULT,
):
    """
    Create EMR cluster for Mage project.
    """
    from mage_ai.services.aws.emr.launcher import create_cluster

    project_path = os.path.abspath(project_path)
    create_cluster(project_path)


if __name__ == '__main__':
    app()



import os

DEFAULT_NAMESPACE = os.getenv('KUBE_NAMESPACE', 'default')
DEFAULT_SERVICE_ACCOUNT_NAME = os.getenv('KUBE_SERVICE_ACCOUNT_NAME', 'default')
DEFAULT_STORAGE_CLASS_NAME = os.getenv('KUBE_STORAGE_CLASS_NAME', 'default')

KUBE_POD_NAME_ENV_VAR = 'HOSTNAME'



from typing import Dict

from mage_ai.ai.constants import LLMUseCase


class Generator:
    @classmethod
    async def generate(self, use_case: LLMUseCase, request: Dict) -> Dict:
        if use_case == LLMUseCase.GENERATE_DOC_FOR_BLOCK:
            from mage_ai.ai.llm_pipeline_wizard import LLMPipelineWizard

            pipeline_uuid = request.get('pipeline_uuid')
            block_uuid = request.get('block_uuid')
            return dict(
                block_doc=await LLMPipelineWizard().async_generate_block_documentation_with_name(
                    pipeline_uuid,
                    block_uuid
                )
            )
        elif use_case == LLMUseCase.GENERATE_DOC_FOR_PIPELINE:
            from mage_ai.ai.llm_pipeline_wizard import LLMPipelineWizard

            pipeline_uuid = request.get('pipeline_uuid')
            return await LLMPipelineWizard().async_generate_pipeline_documentation(
                pipeline_uuid,
            )
        elif use_case == LLMUseCase.GENERATE_BLOCK_WITH_DESCRIPTION:
            from mage_ai.ai.llm_pipeline_wizard import LLMPipelineWizard

            return await LLMPipelineWizard().async_generate_block_with_description(
                request.get('block_description'),
            )

        raise Exception(f'Use case {use_case} is not supported yet.')



from mage_ai.api.operations import constants
from mage_ai.api.presenters.BasePresenter import BasePresenter
from mage_ai.cache.block_action_object.constants import (
    OBJECT_TYPE_BLOCK_FILE,
    OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE,
    OBJECT_TYPE_MAGE_TEMPLATE,
)
from mage_ai.services.search.constants import SEARCH_TYPE_BLOCK_ACTION_OBJECTS
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.strings import remove_extension_from_filename


class SearchResultPresenter(BasePresenter):
    default_attributes = [
        'results',
        'type',
        'uuid',
    ]

    def present(self, **kwargs):
        display_format = kwargs['format']

        if constants.CREATE != display_format:
            return self.model

        results = self.model.get('results', [])
        search_type = self.model.get('type')
        results_transformed = []

        if SEARCH_TYPE_BLOCK_ACTION_OBJECTS == search_type:
            for result in results:
                block_action_object = result.get('block_action_object')
                object_type = result.get('object_type')
                uuid = result.get('uuid')

                block_type = None
                description = block_action_object.get('description')
                language = block_action_object.get('language')
                title = None

                if OBJECT_TYPE_BLOCK_FILE == object_type:
                    block_type = block_action_object.get('type')
                    u = block_action_object.get('uuid')
                    title = ' '.join(list(filter(lambda x: x, [
                        remove_extension_from_filename(u).replace('_', ' ') if u else u,
                    ])))
                elif OBJECT_TYPE_CUSTOM_BLOCK_TEMPLATE == object_type:
                    block_type = block_action_object.get('block_type')
                    template_name = block_action_object.get('name')
                    template_uuid = block_action_object.get('template_uuid')
                    u = template_name or template_uuid
                    title = ' '.join(list(filter(lambda x: x, [
                        remove_extension_from_filename(u).replace('_', ' ') if u else u,
                    ])))
                elif OBJECT_TYPE_MAGE_TEMPLATE == object_type:
                    block_type = block_action_object.get('block_type')
                    u = block_action_object.get('name')
                    title = ' '.join(list(filter(lambda x: x, [
                        remove_extension_from_filename(u).replace('_', ' ') if u else u,
                    ])))

                results_transformed.append(dict(
                    block_type=block_type,
                    description=description,
                    language=language,
                    object_type=object_type,
                    title=title,
                    uuid=uuid,
                ))

        return merge_dict(self.model, dict(
            results=results_transformed,
        ))



from datetime import datetime
from typing import List

from mage_ai.orchestration.db.models.oauth import Oauth2AccessToken, Oauth2Application


def access_tokens_for_provider(provider: str) -> List[Oauth2AccessToken]:
    oauth_client = Oauth2Application.query.filter(
        Oauth2Application.client_id == provider,
    ).first()

    access_tokens = []
    if oauth_client:
        access_tokens = Oauth2AccessToken.query.filter(
            Oauth2AccessToken.expires > datetime.utcnow(),
            Oauth2AccessToken.oauth2_application_id == oauth_client.id,
        )

    return [row for row in access_tokens]



from github import Auth, Github
from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.git import api
from typing import Dict


def pull_request_to_dict(pr) -> Dict:
    return dict(
        body=pr.body,
        created_at=pr.created_at,
        id=pr.id,
        is_merged=pr.is_merged(),
        last_modified=pr.last_modified,
        merged=pr.merged,
        state=pr.state,
        title=pr.title,
        url=pr.html_url,
        user=pr.user.login,
    )


class PullRequestResource(GenericResource):
    @classmethod
    def collection(self, query, meta, user, **kwargs):
        arr = []

        repository = query.get('repository', None)
        if repository:
            repository = repository[0]

            access_token = api.get_access_token_for_user(user)
            if access_token:
                auth = Auth.Token(access_token.token)
                g = Github(auth=auth)
                repo = g.get_repo(repository)
                pulls = repo.get_pulls(
                    direction='desc',
                    sort='created',
                    state='open',
                ).get_page(0)

                for pr in pulls:
                    arr.append(pull_request_to_dict(pr))

        return self.build_result_set(arr, user, **kwargs)

    @classmethod
    def create(self, payload, user, **kwargs):
        error = ApiError.RESOURCE_INVALID

        for key in [
            'base_branch',
            'compare_branch',
            'title',
        ]:
            if key not in payload:
                error.update(dict(message=f'Value for {key} is required but empty.'))
                raise ApiError(error)

        repository = payload.get('repository')
        if not repository:
            error.update(dict(
                message='Repository is empty, ' +
                'please select a repository to create a pull request in.',
            ))
            raise ApiError(error)

        access_token = api.get_access_token_for_user(user)
        if not access_token:
            error.update(dict(
                message='Access token not found, please authenticate with GitHub.',
            ))
            raise ApiError(error)

        auth = Auth.Token(access_token.token)
        g = Github(auth=auth)
        repo = g.get_repo(repository)

        pr = repo.create_pull(
            base=payload.get('base_branch'),
            body=payload.get('body'),
            head=payload.get('compare_branch'),
            title=payload.get('title'),
        )

        return self(pull_request_to_dict(pr), user, **kwargs)



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.global_data_product import GlobalDataProduct
from mage_ai.shared.hash import ignore_keys


class GlobalDataProductResource(GenericResource):
    @classmethod
    def collection(self, query, meta, user, **kwargs):
        return self.build_result_set(
            sorted(GlobalDataProduct.load_all(), key=lambda x: x.uuid),
            user,
            **kwargs,
        )

    @classmethod
    def create(self, payload, user, **kwargs):
        uuid = payload.get('uuid')
        if GlobalDataProduct.get(uuid):
            error = ApiError.RESOURCE_INVALID.copy()
            error.update(dict(message=f'A global data product with UUID {uuid} already exists.'))
            raise ApiError(error)

        model = GlobalDataProduct(uuid, **ignore_keys(payload, ['uuid']))
        model.save()

        return self(model, user, **kwargs)

    @classmethod
    def member(self, pk, user, **kwargs):
        return self(GlobalDataProduct.get(pk), user, **kwargs)

    def delete(self, **kwargs):
        self.model.delete()

    def update(self, payload, **kwargs):
        uuid = payload.get('uuid')
        if self.model.uuid != uuid and GlobalDataProduct.get(uuid):
            error = ApiError.RESOURCE_INVALID.copy()
            error.update(dict(message=f'A global data product with UUID {uuid} already exists.'))
            raise ApiError(error)

        self.model.update(payload)



from mage_ai.shared.hash import get_json_value
from mage_ai.tests.base_test import TestCase


class HashTests(TestCase):
    def test_get_json_value(self):
        self.assertEqual(
            get_json_value('{"k1": "v1", "k2": "v2"}', 'k1'),
            'v1'
        )
        self.assertEqual(
            get_json_value('test_str', 'k1'),
            'test_str'
        )
        self.assertEqual(
            get_json_value('{"k1": "v1", "k2": ["v21", "v22"]}', 'k2[0]'),
            'v21'
        )
        self.assertEqual(
            get_json_value('{"k1": "v1", "k2": {"k21": "v21", "k22": "v22"} }', 'k2.k22'),
            'v22'
        )



import asyncio
import enum
import traceback
import uuid
from datetime import datetime, timedelta, timezone
from math import ceil
from statistics import stdev
from typing import Dict, List

import dateutil.parser
import pytz
from croniter import croniter
from dateutil.relativedelta import relativedelta
from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    String,
    Table,
    or_,
)
from sqlalchemy.orm import joinedload, relationship, validates
from sqlalchemy.sql import func
from sqlalchemy.sql.functions import coalesce

from mage_ai.data_preparation.logging.logger_manager_factory import LoggerManagerFactory
from mage_ai.data_preparation.models.block.utils import (
    get_all_ancestors,
    is_dynamic_block,
    is_dynamic_block_child,
)
from mage_ai.data_preparation.models.constants import ExecutorType, PipelineType
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.triggers import (
    ScheduleInterval,
    ScheduleStatus,
    ScheduleType,
    SettingsConfig,
    Trigger,
)
from mage_ai.data_preparation.variable_manager import get_global_variables
from mage_ai.orchestration.db import db_connection, safe_db_query
from mage_ai.orchestration.db.models.base import Base, BaseModel, classproperty
from mage_ai.settings.repo import get_repo_path
from mage_ai.shared.array import find
from mage_ai.shared.constants import ENV_PROD
from mage_ai.shared.dates import compare
from mage_ai.shared.hash import ignore_keys, index_by, merge_dict
from mage_ai.shared.utils import clean_name
from mage_ai.usage_statistics.logger import UsageStatisticLogger

pipeline_schedule_event_matcher_association_table = Table(
    'pipeline_schedule_event_matcher_association',
    Base.metadata,
    Column('pipeline_schedule_id', ForeignKey('pipeline_schedule.id')),
    Column('event_matcher_id', ForeignKey('event_matcher.id')),
)


class PipelineSchedule(BaseModel):
    name = Column(String(255))
    pipeline_uuid = Column(String(255), index=True)
    schedule_type = Column(Enum(ScheduleType))
    start_time = Column(DateTime(timezone=True), default=None)
    schedule_interval = Column(String(50))
    status = Column(Enum(ScheduleStatus), default=ScheduleStatus.INACTIVE)
    variables = Column(JSON)
    sla = Column(Integer, default=None)  # in seconds
    token = Column(String(255), index=True, default=None)
    repo_path = Column(String(255))
    settings = Column(JSON)

    backfills = relationship('Backfill', back_populates='pipeline_schedule')
    pipeline_runs = relationship('PipelineRun', back_populates='pipeline_schedule')

    event_matchers = relationship(
        'EventMatcher',
        secondary=pipeline_schedule_event_matcher_association_table,
        back_populates='pipeline_schedules'
    )

    @classproperty
    def repo_query(cls):
        return cls.query.filter(
            or_(
                PipelineSchedule.repo_path == get_repo_path(),
                PipelineSchedule.repo_path.is_(None),
            )
        )

    def get_settings(self) -> 'SettingsConfig':
        settings = self.settings if self.settings else dict()
        return SettingsConfig.load(config=settings)

    @property
    def pipeline(self) -> 'Pipeline':
        return Pipeline.get(self.pipeline_uuid)

    @property
    def pipeline_runs_count(self) -> int:
        return len(self.pipeline_runs)

    @validates('schedule_interval')
    def validate_schedule_interval(self, key, schedule_interval):
        if schedule_interval and schedule_interval not in \
                [e.value for e in ScheduleInterval]:
            if not croniter.is_valid(schedule_interval):
                raise ValueError('Cron expression is invalid.')

        return schedule_interval

    @property
    def last_pipeline_run_status(self) -> str:
        if len(self.pipeline_runs) == 0:
            return None
        return sorted(self.pipeline_runs, key=lambda x: x.created_at)[-1].status

    @classmethod
    @safe_db_query
    def active_schedules(self, pipeline_uuids: List[str] = None) -> List['PipelineSchedule']:
        query = self.repo_query.filter(
            self.status == ScheduleStatus.ACTIVE,
        )
        if pipeline_uuids is not None:
            query = query.filter(PipelineSchedule.pipeline_uuid.in_(pipeline_uuids))
        return query.all()

    @classmethod
    def create(self, **kwargs) -> 'PipelineSchedule':
        if 'token' not in kwargs:
            kwargs['token'] = uuid.uuid4().hex
        model = super().create(**kwargs)
        return model

    @classmethod
    @safe_db_query
    def create_or_update(self, trigger_config: Trigger):
        try:
            existing_trigger = PipelineSchedule.repo_query.filter(
                self.name == trigger_config.name,
                self.pipeline_uuid == trigger_config.pipeline_uuid,
            ).one_or_none()
        except Exception:
            traceback.print_exc()
            return

        kwargs = dict(
            name=trigger_config.name,
            pipeline_uuid=trigger_config.pipeline_uuid,
            schedule_type=trigger_config.schedule_type,
            start_time=trigger_config.start_time,
            schedule_interval=trigger_config.schedule_interval,
            status=trigger_config.status,
            variables=trigger_config.variables,
            sla=trigger_config.sla,
            settings=trigger_config.settings,
        )

        if existing_trigger:
            existing_trigger.update(**kwargs)
        else:
            self.create(**kwargs)

    def current_execution_date(self) -> datetime:
        now = datetime.now(timezone.utc)
        current_execution_date = None

        if self.schedule_interval is None:
            return None

        if self.schedule_interval == '@once':
            current_execution_date = now
        elif self.schedule_interval == '@daily':
            current_execution_date = now.replace(second=0, microsecond=0, minute=0, hour=0)
        elif self.schedule_interval == '@hourly':
            current_execution_date = now.replace(second=0, microsecond=0, minute=0)
        elif self.schedule_interval == '@weekly':
            current_execution_date = now.replace(second=0, microsecond=0, minute=0, hour=0) - \
                timedelta(days=now.weekday())
        elif self.schedule_interval == '@monthly':
            current_execution_date = now.replace(second=0, microsecond=0, minute=0, hour=0, day=1)
        else:
            cron_itr = croniter(self.schedule_interval, now)
            current_execution_date = cron_itr.get_prev(datetime)

        # If landing time is enabled, the start_time is used as the date and time the schedule
        # should finish running by.

        # For example, if start_time is 2023-12-01 12:25:00
        # and if the schedule_interval is @daily, then the schedule should finish running by
        # every day by 12:25:00.
        if self.landing_time_enabled() and self.start_time:
            if self.schedule_interval == '@hourly':
                current_execution_date = current_execution_date.replace(
                    minute=self.start_time.minute,
                    second=self.start_time.second,
                )
            elif self.schedule_interval == '@daily':
                current_execution_date = current_execution_date.replace(
                    hour=self.start_time.hour,
                    minute=self.start_time.minute,
                    second=self.start_time.second,
                )
            elif self.schedule_interval == '@weekly':
                current_execution_date = current_execution_date.replace(
                    hour=self.start_time.hour,
                    minute=self.start_time.minute,
                    second=self.start_time.second,
                ) + timedelta(days=self.start_time.weekday())
            elif self.schedule_interval == '@monthly':
                current_execution_date = current_execution_date.replace(
                    hour=self.start_time.hour,
                    minute=self.start_time.minute,
                    second=self.start_time.second,
                ) + timedelta(days=int(self.start_time.strftime('%d')) - 1)

        return current_execution_date

    @safe_db_query
    def should_schedule(self, previous_runtimes: List[int] = None) -> bool:
        now = datetime.now()

        if self.status != ScheduleStatus.ACTIVE:
            return False

        if not self.landing_time_enabled() and \
                self.start_time is not None and \
                compare(now, self.start_time) == -1:

            return False

        try:
            Pipeline.get(self.pipeline_uuid)
        except Exception:
            print(
                f'[WARNING] Pipeline {self.pipeline_uuid} cannot be found '
                + f'for pipeline schedule ID {self.id}.',
            )
            return False

        if self.schedule_interval == '@once':
            pipeline_run_count = len(self.pipeline_runs)
            if pipeline_run_count == 0:
                return True
            executor_count = self.pipeline.executor_count
            # Used by streaming pipeline to launch multiple executors
            if executor_count > 1 and pipeline_run_count < executor_count:
                return True
        else:
            """
            TODO: Implement other schedule interval checks
            """
            current_execution_date = self.current_execution_date()
            if current_execution_date is None:
                return False
            # If there is a pipeline_run with an execution_date the same as the
            # current_execution_date, then don’t schedule
            if not find(
                lambda x: compare(
                    x.execution_date.replace(tzinfo=pytz.UTC),
                    current_execution_date,
                ) == 0,
                self.pipeline_runs
            ):
                if self.landing_time_enabled():
                    if not previous_runtimes or len(previous_runtimes) == 0:
                        return True
                    else:
                        runtime = ceil(sum(previous_runtimes) / len(previous_runtimes))

                        if len(previous_runtimes) >= 2:
                            sd = ceil(stdev(previous_runtimes) / 2)
                        else:
                            sd = 0

                        # This may cause duplicate pipeline runs to be scheduled if
                        # there is more than 1 scheduler running.
                        if datetime.now(timezone.utc) >= current_execution_date - timedelta(
                            seconds=runtime + sd,
                        ):
                            return True
                else:
                    return True
        return False

    def landing_time_enabled(self) -> bool:
        if ScheduleType.TIME != self.schedule_type:
            return False

        if self.schedule_interval not in [
            '@daily',
            '@hourly',
            '@monthly',
            '@weekly',
        ]:
            return False

        return (self.settings or {}).get('landing_time_enabled', False)

    def runtime_history(
        self,
        pipeline_run=None,
        sample_size: int = None,
    ) -> List[float]:
        sample_size_to_use = sample_size if sample_size else 7
        previous_runtimes = []

        if pipeline_run:
            previous_runtimes += (pipeline_run.metrics or {}).get('previous_runtimes', [])

        if len(previous_runtimes) < sample_size_to_use - 1 if pipeline_run else sample_size_to_use:
            pipeline_runs = (
                PipelineRun.
                query.
                filter(
                    PipelineRun.pipeline_schedule_id == self.id,
                    PipelineRun.status == PipelineRun.PipelineRunStatus.COMPLETED,
                )
            )

            if pipeline_run:
                pipeline_runs = (
                    pipeline_runs.
                    filter(
                        PipelineRun.id != pipeline_run.id,
                    )
                )

            pipeline_runs = (
                pipeline_runs.
                order_by(PipelineRun.execution_date.desc()).
                limit(sample_size_to_use).
                all()
            )

            pipeline_runs = sorted(
                pipeline_runs,
                key=lambda pr: pr.execution_date,
                reverse=True,
            )

            for pr in pipeline_runs:
                runtime = (
                    pr.completed_at - pr.created_at
                ).total_seconds()
                previous_runtimes.append(runtime)

        if pipeline_run:
            runtime = (
                pipeline_run.completed_at - pipeline_run.created_at
            ).total_seconds()
            previous_runtimes = [runtime] + previous_runtimes

        return previous_runtimes[:sample_size_to_use]

    def runtime_average(
        self,
        pipeline_run=None,
        sample_size: int = None,
    ) -> float:
        previous_runtimes = self.runtime_history(
            pipeline_run=pipeline_run,
            sample_size=sample_size,
        )

        if len(previous_runtimes) == 0:
            return None

        return round(sum(previous_runtimes) / len(previous_runtimes), 2)


class PipelineRun(BaseModel):
    class PipelineRunStatus(str, enum.Enum):
        INITIAL = 'initial'
        RUNNING = 'running'
        COMPLETED = 'completed'
        FAILED = 'failed'
        CANCELLED = 'cancelled'

    pipeline_schedule_id = Column(Integer, ForeignKey('pipeline_schedule.id'), index=True)
    pipeline_uuid = Column(String(255), index=True)
    execution_date = Column(DateTime(timezone=True), index=True)
    status = Column(Enum(PipelineRunStatus), default=PipelineRunStatus.INITIAL, index=True)
    completed_at = Column(DateTime(timezone=True))
    variables = Column(JSON)
    passed_sla = Column(Boolean, default=False)
    event_variables = Column(JSON)
    metrics = Column(JSON)
    backfill_id = Column(Integer, ForeignKey('backfill.id'), index=True)
    executor_type = Column(Enum(ExecutorType), default=ExecutorType.LOCAL_PYTHON)

    pipeline_schedule = relationship(PipelineSchedule, back_populates='pipeline_runs')
    block_runs = relationship('BlockRun', back_populates='pipeline_run')
    backfill = relationship('Backfill', back_populates='pipeline_runs')

    def __repr__(self):
        return f'PipelineRun(id={self.id}, pipeline_uuid={self.pipeline_uuid},'\
               f' execution_date={self.execution_date})'

    @property
    def block_runs_count(self) -> int:
        return len(self.block_runs)

    @property
    def execution_partition(self) -> str:
        if self.variables and self.variables.get('execution_partition'):
            return self.variables.get('execution_partition')
        if self.execution_date is None:
            return str(self.pipeline_schedule_id)
        else:
            return '/'.join([
                        str(self.pipeline_schedule_id),
                        self.execution_date.strftime(format='%Y%m%dT%H%M%S'),
                    ])

    @property
    def initial_block_runs(self) -> List['BlockRun']:
        return [b for b in self.block_runs
                if b.status == BlockRun.BlockRunStatus.INITIAL]

    @property
    def queued_or_running_block_runs(self) -> List['BlockRun']:
        return [
            b for b in self.block_runs
            if b.status in [
                BlockRun.BlockRunStatus.QUEUED,
                BlockRun.BlockRunStatus.RUNNING,
            ]
        ]

    @property
    def completed_block_runs(self) -> List['BlockRun']:
        return [b for b in self.block_runs
                if b.status == BlockRun.BlockRunStatus.COMPLETED]

    @property
    def failed_block_runs(self) -> List['BlockRun']:
        return [b for b in self.block_runs
                if b.status == BlockRun.BlockRunStatus.FAILED]

    @property
    def pipeline(self) -> 'Pipeline':
        return Pipeline.get(self.pipeline_uuid)

    @property
    def pipeline_type(self) -> PipelineType:
        pipeline = Pipeline.get(self.pipeline_uuid, check_if_exists=True)

        return self.pipeline.type if pipeline is not None else None

    @property
    def logs(self):
        return LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=self.pipeline_uuid,
            partition=self.execution_partition,
            repo_config=self.pipeline.repo_config,
        ).get_logs()

    async def logs_async(self):
        return await LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=self.pipeline_uuid,
            partition=self.execution_partition,
            repo_config=self.pipeline.repo_config,
        ).get_logs_async()

    @property
    def pipeline_schedule_name(self):
        return self.pipeline_schedule.name

    @property
    def pipeline_schedule_token(self):
        return self.pipeline_schedule.token

    @property
    def pipeline_schedule_type(self):
        return self.pipeline_schedule.schedule_type

    @classmethod
    @safe_db_query
    def active_runs_for_pipelines(
        self,
        pipeline_uuids: List[str],
        include_block_runs: bool = False,
    ) -> List['PipelineRun']:
        # Filter by schedules because pipelines across repos can potentially
        # have the same uuid
        repo_schedules = PipelineSchedule.repo_query.filter(
            PipelineSchedule.pipeline_uuid.in_(pipeline_uuids)
        ).all()
        schedule_ids = [s.id for s in repo_schedules]
        query = self.query.filter(
            self.status == self.PipelineRunStatus.RUNNING,
            self.pipeline_schedule_id.in_(schedule_ids),
        )
        if include_block_runs:
            query = query.options(joinedload(PipelineRun.block_runs))
        return query.all()

    @classmethod
    def create(self, create_block_runs: bool = True, **kwargs) -> 'PipelineRun':
        pipeline_run = super().create(**kwargs)
        pipeline_uuid = kwargs.get('pipeline_uuid')
        if pipeline_uuid is not None and create_block_runs:
            pipeline_run.create_block_runs()

        return pipeline_run

    @classmethod
    @safe_db_query
    def in_progress_runs(
        self,
        pipeline_schedules: List[int],
    ):
        return self.query.filter(
            PipelineRun.pipeline_schedule_id.in_(pipeline_schedules),
            PipelineRun.status.in_([
                self.PipelineRunStatus.INITIAL,
                self.PipelineRunStatus.RUNNING,
            ]),
            (coalesce(PipelineRun.passed_sla, False) == False),  # noqa: E712
        ).all()

    @safe_db_query
    def complete(self):
        self.update(
            completed_at=datetime.now(),
            status=self.PipelineRunStatus.COMPLETED,
        )

        asyncio.run(UsageStatisticLogger().pipeline_runs_impression(
            lambda: self.query.filter(self.status == self.PipelineRunStatus.COMPLETED).count(),
        ))

    @safe_db_query
    def create_block_run(
        self,
        block_uuid: str,
        skip_if_exists: bool = False,
        **kwargs,
    ) -> 'BlockRun':
        if skip_if_exists:
            br = BlockRun.get(
                pipeline_run_id=self.id,
                block_uuid=block_uuid,
            )
            if br is not None:
                return br
        return BlockRun.create(
            block_uuid=block_uuid,
            pipeline_run_id=self.id,
            **kwargs,
        )

    def create_block_runs(self) -> List['BlockRun']:
        pipeline = self.pipeline
        blocks = pipeline.get_executable_blocks()

        arr = []
        for block in blocks:
            ancestors = get_all_ancestors(block)
            if len(block.upstream_blocks) == 0 or not find(is_dynamic_block, ancestors):
                arr.append(block)

        block_uuids = []

        for block in arr:
            block_uuid = block.uuid
            if block.replicated_block:
                replicated_block = pipeline.get_block(block.replicated_block)
                if replicated_block:
                    block_uuid = f'{block.uuid}:{replicated_block.uuid}'
                else:
                    raise Exception(
                        f'Replicated block {block.replicated_block} ' +
                        f'does not exist in pipeline {pipeline.uuid}.',
                    )

            block_uuids.append(block_uuid)

        return [self.create_block_run(block_uuid) for block_uuid in block_uuids]

    def any_blocks_failed(self) -> bool:
        return any(
            b.status == BlockRun.BlockRunStatus.FAILED
            for b in self.block_runs
        )

    def all_blocks_completed(self, include_failed_blocks: bool = False) -> bool:
        statuses = [
            BlockRun.BlockRunStatus.COMPLETED,
            BlockRun.BlockRunStatus.CONDITION_FAILED,
        ]
        if include_failed_blocks:
            statuses.extend([
                BlockRun.BlockRunStatus.FAILED,
                BlockRun.BlockRunStatus.UPSTREAM_FAILED,
            ])
        return all(b.status in statuses for b in self.block_runs)

    def get_variables(self, extra_variables: Dict = None) -> Dict:
        if extra_variables is None:
            extra_variables = dict()

        pipeline_run_variables = self.variables or {}
        event_variables = self.event_variables or {}

        variables = merge_dict(
            merge_dict(
                get_global_variables(self.pipeline_uuid) or dict(),
                self.pipeline_schedule.variables or dict(),
            ),
            pipeline_run_variables,
        )

        # For backwards compatibility
        for k, v in event_variables.items():
            if k not in variables:
                variables[k] = v

        if self.execution_date:
            variables['ds'] = self.execution_date.strftime('%Y-%m-%d')
            variables['hr'] = self.execution_date.strftime('%H')

        variables['env'] = ENV_PROD
        variables['event'] = merge_dict(variables.get('event', {}), event_variables)
        variables['execution_date'] = self.execution_date
        variables['execution_partition'] = self.execution_partition

        interval_end_datetime = variables.get('interval_end_datetime')
        interval_seconds = variables.get('interval_seconds')
        interval_start_datetime = variables.get('interval_start_datetime')
        interval_start_datetime_previous = variables.get('interval_start_datetime_previous')

        if interval_end_datetime or \
                interval_seconds or \
                interval_start_datetime or \
                interval_start_datetime_previous:
            if interval_end_datetime:
                try:
                    variables['interval_end_datetime'] = dateutil.parser.parse(
                        interval_end_datetime,
                    )
                except Exception as err:
                    print(f'[ERROR] PipelineRun.get_variables: {err}')

            if interval_start_datetime:
                try:
                    variables['interval_start_datetime'] = dateutil.parser.parse(
                        interval_start_datetime,
                    )
                except Exception as err:
                    print(f'[ERROR] PipelineRun.get_variables: {err}')

            if interval_start_datetime_previous:
                try:
                    variables['interval_start_datetime_previous'] = dateutil.parser.parse(
                        interval_start_datetime_previous,
                    )
                except Exception as err:
                    print(f'[ERROR] PipelineRun.get_variables: {err}')
        elif self.execution_date and ScheduleType.TIME == self.pipeline_schedule.schedule_type:
            interval_end_datetime = None
            interval_seconds = None
            interval_start_datetime = self.execution_date
            interval_start_datetime_previous = None

            if ScheduleInterval.DAILY == self.pipeline_schedule.schedule_interval:
                interval_seconds = 60 * 60 * 24
            elif ScheduleInterval.HOURLY == self.pipeline_schedule.schedule_interval:
                interval_seconds = 60 * 60 * 1
            elif ScheduleInterval.MONTHLY == self.pipeline_schedule.schedule_interval:
                interval_end_datetime = interval_start_datetime + relativedelta(months=1)
                interval_seconds = (
                    interval_end_datetime.timestamp() - interval_start_datetime.timestamp()
                )
            elif ScheduleInterval.WEEKLY == self.pipeline_schedule.schedule_interval:
                interval_seconds = 60 * 60 * 24 * 7

            if interval_seconds and not interval_end_datetime:
                interval_end_datetime = interval_start_datetime + timedelta(
                    seconds=interval_seconds,
                )

            if interval_seconds and interval_start_datetime:
                interval_start_datetime_previous = interval_start_datetime - timedelta(
                    seconds=interval_seconds,
                )

            variables['interval_end_datetime'] = interval_end_datetime
            variables['interval_seconds'] = interval_seconds
            variables['interval_start_datetime'] = interval_start_datetime
            variables['interval_start_datetime_previous'] = interval_start_datetime_previous

        variables.update(extra_variables)

        return variables


class BlockRun(BaseModel):
    class BlockRunStatus(str, enum.Enum):
        INITIAL = 'initial'
        QUEUED = 'queued'
        RUNNING = 'running'
        COMPLETED = 'completed'
        FAILED = 'failed'
        CANCELLED = 'cancelled'
        UPSTREAM_FAILED = 'upstream_failed'
        CONDITION_FAILED = 'condition_failed'

    pipeline_run_id = Column(Integer, ForeignKey('pipeline_run.id'), index=True)
    block_uuid = Column(String(255))
    status = Column(Enum(BlockRunStatus), default=BlockRunStatus.INITIAL)
    started_at = Column(DateTime(timezone=True))
    completed_at = Column(DateTime(timezone=True))
    metrics = Column(JSON)

    pipeline_run = relationship(PipelineRun, back_populates='block_runs')

    @property
    def logs(self):
        pipeline = Pipeline.get(self.pipeline_run.pipeline_uuid)
        return LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=pipeline.uuid,
            block_uuid=clean_name(self.block_uuid),
            partition=self.pipeline_run.execution_partition,
            repo_config=pipeline.repo_config,
        ).get_logs()

    async def logs_async(self):
        pipeline = await Pipeline.get_async(self.pipeline_run.pipeline_uuid)
        return await LoggerManagerFactory.get_logger_manager(
            pipeline_uuid=pipeline.uuid,
            block_uuid=clean_name(self.block_uuid),
            partition=self.pipeline_run.execution_partition,
            repo_config=pipeline.repo_config,
        ).get_logs_async()

    @classmethod
    @safe_db_query
    def batch_update_status(self, block_run_ids: List[int], status):
        BlockRun.query.filter(BlockRun.id.in_(block_run_ids)).update({
            BlockRun.status: status
        }, synchronize_session=False)
        db_connection.session.commit()

    @classmethod
    @safe_db_query
    def get(self, pipeline_run_id: int = None, block_uuid: str = None) -> 'BlockRun':
        block_runs = self.query.filter(
            BlockRun.pipeline_run_id == pipeline_run_id,
            BlockRun.block_uuid == block_uuid,
        ).all()
        if len(block_runs) > 0:
            return block_runs[0]
        return None

    def get_outputs(self, sample_count: int = None) -> List[Dict]:
        pipeline = Pipeline.get(self.pipeline_run.pipeline_uuid)
        block = pipeline.get_block(self.block_uuid)
        block_uuid = self.block_uuid

        # The block_run’s block_uuid for replicated blocks will be in this format:
        # [block_uuid]:[replicated_block_uuid]
        # We need to use the original block_uuid to get the proper output.

        # Block runs for dynamic child blocks will have the following block UUID:
        # [block.uuid]:[index]
        # Don’t use the original UUID even if the block is a replica because it will get rid of
        # the dynamic child block index.
        if block.replicated_block and not is_dynamic_block_child(block):
            block_uuid = block.uuid

        return block.get_outputs(
            execution_partition=self.pipeline_run.execution_partition,
            sample_count=sample_count,
            block_uuid=block_uuid,
        )


class EventMatcher(BaseModel):
    class EventType(str, enum.Enum):
        AWS_EVENT = 'aws_event'

    event_type = Column(Enum(EventType), default=EventType.AWS_EVENT)
    name = Column(String(255))
    pattern = Column(JSON)

    pipeline_schedules = relationship(
        'PipelineSchedule',
        secondary=pipeline_schedule_event_matcher_association_table,
        back_populates='event_matchers',
    )

    def __repr__(self):
        return f'EventMatcher(id={self.id}, name={self.name}, pattern={self.pattern})'

    @classmethod
    def active_event_matchers(self) -> List['EventMatcher']:
        return self.query.filter(
            EventMatcher.pipeline_schedules.any(
                PipelineSchedule.status == ScheduleStatus.ACTIVE
            )
        ).all()

    @classmethod
    def upsert_batch(self, event_matchers_payload):
        keys_to_ignore = [
            'created_at',
            'id',
            'updated_at',
        ]

        new_arr = []
        existing_arr = []

        pipeline_schedule_ids = []
        for payload in event_matchers_payload:
            pipeline_schedule_ids += payload.get('pipeline_schedule_ids', [])
            if payload.get('id'):
                existing_arr.append(payload)
            else:
                new_arr.append(payload)

        pipeline_schedules_by_id = index_by(
            lambda x: x.id,
            PipelineSchedule.query.filter(
                PipelineSchedule.id.in_(pipeline_schedule_ids),
            ).all(),
        )

        event_matchers_and_pipeline_schedule_ids = []
        event_matchers_by_id = index_by(
            lambda x: x.id,
            self.query.filter(
                self.id.in_([p['id'] for p in existing_arr]),
            ).all(),
        )
        for payload in existing_arr:
            ids = payload.pop('pipeline_schedule_ids', None)
            event_matcher = event_matchers_by_id[payload['id']]
            event_matcher.update(**ignore_keys(payload, keys_to_ignore))
            event_matchers_and_pipeline_schedule_ids.append((event_matcher, ids))

        for payload in new_arr:
            ids = payload.pop('pipeline_schedule_ids', None)
            event_matcher = self.create(**ignore_keys(payload, keys_to_ignore))
            event_matchers_and_pipeline_schedule_ids.append((event_matcher, ids))

        for event_matcher, ids in event_matchers_and_pipeline_schedule_ids:
            if ids is not None:
                ps = [pipeline_schedules_by_id[i] for i in [int(i) for i in ids]]
                event_matcher.update(pipeline_schedules=ps)

            if event_matcher.event_type == EventMatcher.EventType.AWS_EVENT:
                from mage_ai.services.aws.events.events import update_event_rule_targets

                # For AWS event, update related AWS infra (add trigger to lambda function)
                update_event_rule_targets(event_matcher.name)

        return [t[0] for t in event_matchers_and_pipeline_schedule_ids]

    def active_pipeline_schedules(self) -> List[PipelineSchedule]:
        return [p for p in self.pipeline_schedules
                if p.status == ScheduleStatus.ACTIVE]

    def match(self, config: Dict) -> bool:
        def __match_dict(sub_pattern, sub_config):
            if type(sub_pattern) is not dict or type(sub_config) is not dict:
                return False
            for k in sub_pattern.keys():
                if k not in sub_config:
                    return False
                v = sub_pattern[k]
                if type(v) is list:
                    if sub_config[k] not in v:
                        return False
                elif not __match_dict(v, sub_config[k]):
                    return False
            return True
        return __match_dict(self.pattern, config)


class Backfill(BaseModel):
    class IntervalType(str, enum.Enum):
        SECOND = 'second'
        MINUTE = 'minute'
        HOUR = 'hour'
        DAY = 'day'
        WEEK = 'week'
        MONTH = 'month'
        YEAR = 'year'
        CUSTOM = 'custom'

    class Status(str, enum.Enum):
        INITIAL = 'initial'
        RUNNING = 'running'
        COMPLETED = 'completed'
        FAILED = 'failed'
        CANCELLED = 'cancelled'

    block_uuid = Column(String(255), default=None)
    completed_at = Column(DateTime(timezone=True), default=None)
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    end_datetime = Column(DateTime(timezone=True), default=None)
    failed_at = Column(DateTime(timezone=True), default=None)
    interval_type = Column(Enum(IntervalType), default=None)
    interval_units = Column(Integer, default=None)
    metrics = Column(JSON)
    name = Column(String(255))
    pipeline_schedule = relationship(PipelineSchedule, back_populates='backfills')
    pipeline_schedule_id = Column(Integer, ForeignKey('pipeline_schedule.id'))
    pipeline_uuid = Column(String(255))
    start_datetime = Column(DateTime(timezone=True), default=None)
    started_at = Column(DateTime(timezone=True), default=None)
    status = Column(Enum(Status), default=None)
    updated_at = Column(DateTime(timezone=True), server_default=func.now(), onupdate=func.now())
    variables = Column(JSON, default=None)

    pipeline_runs = relationship('PipelineRun', back_populates='backfill')

    @classmethod
    @safe_db_query
    def filter(self, pipeline_schedule_ids: List = None):
        if pipeline_schedule_ids is not None:
            return Backfill.query.filter(
                Backfill.pipeline_schedule_id.in_(pipeline_schedule_ids),
            )
        return []



import traceback

import redis


def init_redis_client(redis_url):
    if not redis_url:
        return None
    try:
        redis_client = redis.Redis.from_url(url=redis_url, decode_responses=True)
        redis_client.ping()
    except Exception:
        traceback.print_exc()
        redis_client = None
    return redis_client



from mage_ai.ai.generator import Generator
from mage_ai.api.resources.GenericResource import GenericResource


class LlmResource(GenericResource):
    @classmethod
    async def create(self, payload, user, **kwargs):
        response = await Generator.generate(payload.get('use_case'), payload.get('request'))

        return self(dict(
            use_case=payload.get('use_case'),
            response=response,
        ), user, **kwargs)



CUSTOM_TEMPLATES_DIRECTORY = 'custom_templates'
CUSTOM_TEMPLATES_DIRECTORY_ENVIRONMENT_VARIABLE = 'CUSTOM_TEMPLATES_DIRECTORY'

DIRECTORY_FOR_BLOCK_TEMPLATES = 'blocks'
DIRECTORY_FOR_PIPELINE_TEMPLATES = 'pipelines'

METADATA_FILENAME_WITH_EXTENSION = 'metadata.yaml'



from unittest.mock import patch

from mage_ai.streaming.sinks.azure_data_lake import AzureDataLakeSink
from mage_ai.tests.base_test import TestCase


class AzureDLSTests(TestCase):
    def test_init(self):
        with patch.object(AzureDataLakeSink, 'init_client') as mock_init_client:
            AzureDataLakeSink(dict(
                table_uri='table_uri',
                account_name='account_name',
                access_key='access_key',
            ))
            mock_init_client.assert_called_once()

    def test_init_invalid_config(self):
        with patch.object(AzureDataLakeSink, 'init_client') as mock_init_client:
            with self.assertRaises(Exception) as context:
                AzureDataLakeSink(dict(
                    account_name='account_name',
                    access_key='access_key',
                ))
            self.assertTrue(
                '__init__() missing 1 required positional argument: \'table_uri\''
                in str(context.exception),
            )
            self.assertEqual(mock_init_client.call_count, 0)
